# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import join, exists, basename

from tsfpga.system_utils import load_python_module
from tsfpga.module import iterate_module_folders


class FPGAProjectList:

    def __init__(self, modules_folders):
        self._get_projects(modules_folders)

    @staticmethod
    def _get_project_objects(folder, module_name):
        project_file = join(folder, "project_" + module_name + ".py")

        if exists(project_file):
            return load_python_module(project_file).get_projects()
        return []

    def _get_projects(self, modules_folders):
        self.projects = []

        for module_folder in iterate_module_folders(modules_folders):
            module_name = basename(module_folder)
            self.projects += self._get_project_objects(module_folder, module_name)

    def get(self, project_name):
        for project in self.projects:
            if project.name == project_name:
                return project

    def names(self):
        names = []
        for project in self.projects:
            names.append(project.name)
        return names

    def __str__(self):
        return "\n\n".join(["%s" % project for project in self.projects])



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import dirname, join
import pytest
import unittest

from tsfpga.constraint import Constraint
from tsfpga.hdl_file import HdlFile
from tsfpga.system_utils import create_file, delete


THIS_DIR = dirname(__file__)


class TestConstraint(unittest.TestCase):

    _modules_folder = join(THIS_DIR, "modules_for_test")

    def setUp(self):
        delete(self._modules_folder)
        self.file = create_file(join(self._modules_folder, "a", "scoped_constraints", "apa.tcl"))

    def test_constraint(self):
        constraint = Constraint(self.file)
        constraint.validate_scoped_entity([])

        assert constraint.ref is None
        assert constraint.used_in == "all"

        constraint = Constraint(self.file, used_in="impl")
        assert constraint.used_in == "impl"

    def test_scoped_constraint(self):
        constraint = Constraint(self.file, scoped_constraint=True)
        assert constraint.ref == "apa"

        source_files = [HdlFile(join(self._modules_folder, "a", "apa.vhd"))]
        constraint.validate_scoped_entity(source_files)

    def test_matching_entity_not_existing_should_raise_exception(self):
        constraint = Constraint(self.file, scoped_constraint=True)
        with pytest.raises(FileNotFoundError) as exception_info:
            constraint.validate_scoped_entity([])
        assert str(exception_info.value).startswith("Could not find a matching entity file")



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import dirname, join
import unittest

from tsfpga.system_utils import create_file, create_directory, delete
from tsfpga.fpga_project_list import FPGAProjectList


THIS_DIR = dirname(__file__)


class TestFPGAProjectList(unittest.TestCase):

    _modules_folder = join(THIS_DIR, "modules_for_test")
    _modules_folders = [_modules_folder]

    def setUp(self):
        delete(self._modules_folder)
        create_directory(join(self._modules_folder, "a"))
        create_directory(join(self._modules_folder, "b"))
        create_directory(join(self._modules_folder, "c"))

    def test_get_projects(self):
        module_file_content = """
class Dummy:

    def __init__(self):
        self.name = \"{name}\"

def get_projects():
    return [Dummy()]
"""

        create_file(join(self._modules_folder, "a", "project_a.py"), module_file_content.format(name="a"))
        create_file(join(self._modules_folder, "b", "project_b.py"), module_file_content.format(name="b"))

        projects = FPGAProjectList(self._modules_folders)

        assert len(projects.names()) == 2
        assert projects.get("a").name == "a"
        assert projects.get("b").name == "b"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import basename, splitext


class Constraint:
    """
    Class for handling a constraint file.

    Can handle the regular global constraint files as well as scoped constraints.
    For the latter to work the constraint file name must be the same as the module file name, which must the same as entity name.
    """

    def __init__(self, file, used_in="all", scoped_constraint=False, processing_order="normal"):
        self.file = file
        self.used_in = used_in
        self.ref = splitext(basename(file))[0] if scoped_constraint else None
        self.processing_order = processing_order.lower()

        assert self.used_in in ["all", "synth", "impl"], self.used_in
        assert self.processing_order in ["early", "normal", "late"], self.processing_order

    def validate_scoped_entity(self, source_files):
        """
        Make sure that a matching entity file exists in case this is a scoped constraint. The list of source files should be
        the synthesis files for the module that this constraint belongs to.
        """
        if self.ref is not None:
            entity_file_name = self.ref + ".vhd"
            if not any([source_file.filename.endswith(entity_file_name)] for source_file in source_files):
                raise FileNotFoundError(f"Could not find a matching entity file {entity_file_name} for constraint file {self.file}")
        return True

    def __str__(self):
        return self.file



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import join, exists
import hashlib

from tsfpga.system_utils import create_file, delete
from tsfpga.vivado_project import VivadoProject


class VivadoIpCores:

    """
    Handle a list of IP core sources. Has a mechanism to detect whether a regenerate of IP files is needed.
    """

    _project_name = "vivado_ip_project"

    def __init__(self, modules, project_path, part_name="xc7z020clg400-1"):
        self._project_folder = join(project_path, self._project_name)
        self._part_name = part_name
        self._hash_file = join(self._project_folder, "ip_files_hash.txt")

        self._setup(modules)

    @property
    def compile_order_file(self):
        """
        Path to the generated compile order file.
        """
        return join(self._project_folder, "compile_order.txt")

    @property
    def vivado_project_sources_directory(self):
        """
        Path to the "sources" directory of the Vivado project.
        """
        return join(self._project_folder, "vivado_ip_project.srcs", "sources_1")

    @property
    def vivado_project_file(self):
        """
        Path to the Vivado project file.
        """
        return self._vivado_project.project_file(self._project_folder)

    def create_vivado_project(self):
        """
        Create IP core Vivado project.
        """
        print(f"Creating IP core project in {self._project_folder}")
        delete(self._project_folder)
        self._vivado_project.create(self._project_folder)
        self._save_hash()

    def create_vivado_project_if_needed(self):
        """
        Create IP core Vivado project if anything has changed since last time this was run.
        """
        if self._should_create():
            self.create_vivado_project()
            return True

        return False

    def _setup(self, modules):
        ip_tcl_files = []
        for module in modules:
            ip_tcl_files += module.get_ip_core_files()

        self._vivado_project = VivadoProject(
            name=self._project_name,
            modules=[],
            part=self._part_name,
            tcl_sources=ip_tcl_files
        )

        self._hash = self._calculate_hash(ip_tcl_files)

    @staticmethod
    def _calculate_hash(files):
        ip_hash = hashlib.md5()
        for file in files:
            ip_hash.update(file.encode())
            with open(file) as file_handle:
                ip_hash.update(file_handle.read().encode())

        return ip_hash.hexdigest()

    def _save_hash(self):
        create_file(self._hash_file, self._hash)

    def _should_create(self):
        """
        Return True if a Vivado project create is needed, i.e. if anything has changed.
        """
        if not (exists(self._hash_file) and exists(self.compile_order_file)):
            return True

        with open(self._hash_file) as file_handle:
            saved_hash = file_handle.read()
        if saved_hash != self._hash:
            return True

        return False



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from unittest import mock, TestCase
from os.path import dirname, join
import pytest

from tsfpga.module import BaseModule, get_modules
from tsfpga.system_utils import create_file, create_directory, delete


THIS_DIR = dirname(__file__)


def test_file_list_filtering(tmpdir):
    module_name = "zebra"
    path = join(tmpdir, module_name)

    create_directory(join(path, "folder_should_not_be_included"))
    create_file(join(path, "should_not_be_included.apa"))

    synth_files = [create_file(join(path, "syn.v")),
                   create_file(join(path, "rtl", "syn.v")),
                   create_file(join(path, "src", "syn.vhd")),
                   create_file(join(path, "hdl", "rtl", "syn.vhd")),
                   create_file(join(path, "hdl", "package", "syn.vhd"))]

    test_files = [create_file(join(path, "test", "test.v")),
                  create_file(join(path, "rtl", "tb", "test.vhd"))]

    sim_files = [create_file(join(path, "sim", "sim.vhd"))]

    my_module = BaseModule(path)

    files = [file.filename for file in my_module.get_synthesis_files()]
    assert set(files) == set(synth_files)

    files = [file.filename for file in my_module.get_simulation_files()]
    assert set(files) == set(synth_files + test_files + sim_files)

    files = [file.filename for file in my_module.get_simulation_files(include_tests=False)]
    assert set(files) == set(synth_files + sim_files)


def test_scoped_constraints(tmpdir):
    module_path = join(tmpdir, "apa")
    create_file(join(module_path, "src", "hest.vhd"))
    create_file(join(module_path, "scoped_constraints", "hest.tcl"))

    my_module = BaseModule(module_path)
    scoped_constraints = my_module.get_scoped_constraints()
    assert len(scoped_constraints) == 1
    assert scoped_constraints[0].ref == "hest"


def test_scoped_constraint_entity_not_existing_should_raise_error(tmpdir):
    module_path = join(tmpdir, "apa")
    create_file(join(module_path, "scoped_constraints", "hest.tcl"))

    module = BaseModule(module_path)
    with pytest.raises(FileNotFoundError) as exception_info:
        module.get_scoped_constraints()
    assert str(exception_info.value).startswith("Could not find a matching entity file")


class TestGetModules(TestCase):

    _modules_folder = join(THIS_DIR, "modules_for_test")
    _modules_folders = [_modules_folder]

    def setUp(self):
        delete(self._modules_folder)
        create_directory(join(self._modules_folder, "a"))
        create_directory(join(self._modules_folder, "b"))
        create_directory(join(self._modules_folder, "c"))

    def test_name_filtering_include(self):
        modules = get_modules(self._modules_folders, names_include=["a", "b"])
        assert set(module.name for module in modules) == set(["a", "b"])

    def test_name_filtering_avoid(self):
        modules = get_modules(self._modules_folders, names_avoid=["a", "b"])
        assert set(module.name for module in modules) == set(["c"])

    def test_name_filtering_include_and_avoid(self):
        modules = get_modules(self._modules_folders,
                              names_include=["a", "c"],
                              names_avoid=["b", "c"])
        assert set(module.name for module in modules) == set(["a"])

    def test_library_name_does_not_have_lib_suffix(self):
        modules = get_modules(self._modules_folders)
        assert set(module.library_name for module in modules) == set(["a", "b", "c"])

    def test_library_name_has_lib_suffix(self):
        modules = get_modules(self._modules_folders, library_name_has_lib_suffix=True)
        assert set(module.library_name for module in modules) == set(["a_lib", "b_lib", "c_lib"])

    def test_stray_file_can_exist_in_modules_folder_without_error(self):
        create_file(join(self._modules_folder, "text_file.txt"))
        modules = get_modules(self._modules_folders)
        assert len(modules) == 3

    def test_local_override_of_module_type(self):
        module_file_content = """
from tsfpga.module import BaseModule

class Module(BaseModule):

    def id(self):
        return """

        create_file(join(self._modules_folder, "a", "module_a.py"), module_file_content + "\"a\"")
        create_file(join(self._modules_folder, "b", "module_b.py"), module_file_content + "\"b\"")

        modules = get_modules(self._modules_folders)

        assert len(modules) == 3
        for module in modules:
            if module.name == "a":
                assert module.id() == "a"
            elif module.name == "b":
                assert module.id() == "b"
            elif module.name == "c":
                assert isinstance(module, BaseModule)
            else:
                assert False

    @mock.patch("tsfpga.module.from_json", autospec=True)
    def test_register_object_creation_synthesis(self, from_json):
        json_file = create_file(join(self._modules_folder, "a", "a_regs.json"))

        module = get_modules(self._modules_folders, names_include=["a"])[0]
        module.get_synthesis_files()
        module.get_synthesis_files()

        from_json.assert_called_once_with("a", json_file, mock.ANY)

    @mock.patch("tsfpga.module.from_json", autospec=True)
    def test_register_object_creation_simulation(self, from_json):
        json_file = create_file(join(self._modules_folder, "a", "a_regs.json"))

        module = get_modules(self._modules_folders, names_include=["a"])[0]
        module.get_simulation_files()
        module.get_simulation_files()

        from_json.assert_called_once_with("a", json_file, mock.ANY)

    @mock.patch("tsfpga.module.from_json", autospec=True)
    def test_register_object_creation_mixed(self, from_json):
        json_file = create_file(join(self._modules_folder, "a", "a_regs.json"))

        module = get_modules(self._modules_folders, names_include=["a"])[0]
        module.get_synthesis_files()
        module.get_simulation_files()

        from_json.assert_called_once_with("a", json_file, mock.ANY)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import join, dirname
import unittest

from tsfpga.system_utils import create_file, delete
from tsfpga.test import file_equals
from tsfpga.fix_lint import fix_trailing_whitespace, fix_tabs


THIS_DIR = dirname(__file__)


class TestFixLint(unittest.TestCase):

    file = join(THIS_DIR, "temp_file_for_test.txt")

    def setUp(self):
        delete(self.file)

    def test_fix_trailing_whitespace(self):
        data = "Apa \nhest    \nzebra"
        data_fixed = "Apa\nhest\nzebra"

        create_file(self.file, data)
        fix_trailing_whitespace(self.file)
        assert file_equals(self.file, data_fixed)

    def test_fix_tabs(self):
        data = "Apa\thest \t zebra"
        data_fixed = "Apa hest   zebra"
        create_file(self.file, data)
        fix_tabs(self.file, tab_width=1)
        assert file_equals(self.file, data_fixed)

        data_fixed = "Apa  hest    zebra"
        create_file(self.file, data)
        fix_tabs(self.file, tab_width=2)
        assert file_equals(self.file, data_fixed)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import dirname, join
from unittest import mock, TestCase

from tsfpga.module import get_modules
from tsfpga.system_utils import create_file, delete
from tsfpga.vivado_ip_cores import VivadoIpCores


THIS_DIR = dirname(__file__)


class TestVivadoIpCores(TestCase):

    project_folder = join(THIS_DIR, "ip_project")
    modules_folder = join(THIS_DIR, "modules")

    def setUp(self):
        delete(self.project_folder)
        delete(self.modules_folder)

        self.apa_tcl = create_file(join(self.modules_folder, "apa", "ip_cores", "apa.tcl"), "apa")
        self.hest_tcl = create_file(join(self.modules_folder, "hest", "ip_cores", "hest.tcl"), "hest")

        modules = get_modules([self.modules_folder])
        self.vivado_ip_cores = VivadoIpCores(modules, self.project_folder)

        # Create inital hash and (empty) compile order file
        self.vivado_ip_cores._save_hash()  # pylint: disable=protected-access
        create_file(self.vivado_ip_cores.compile_order_file)

    @mock.patch("tsfpga.vivado_ip_cores.VivadoProject.create", autospec=True)
    def test_should_not_recreate(self, create):
        assert not self.vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_not_called()

    @mock.patch("tsfpga.vivado_ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_compile_order_file_is_missing(self, create):
        delete(self.vivado_ip_cores.compile_order_file)
        assert self.vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()

    @mock.patch("tsfpga.vivado_ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_hash_file_is_missing(self, create):
        delete(self.vivado_ip_cores._hash_file)  # pylint: disable=protected-access
        assert self.vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()

    @mock.patch("tsfpga.vivado_ip_cores.VivadoProject.create", autospec=True)
    def test_should_not_recreate_if_nothing_is_changed(self, create):
        # This test shows that the pattern used in the upcoming tests:
        #   change something -> get modules -> create new VivadoIpCores object
        # should not result in a recreate unless we actually change something.
        modules = get_modules([self.modules_folder])
        vivado_ip_cores = VivadoIpCores(modules, self.project_folder)

        assert not vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_not_called()

    @mock.patch("tsfpga.vivado_ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_ip_core_file_is_added(self, create):
        create_file(join(self.modules_folder, "zebra", "ip_cores", "zebra.tcl"), "zebra")
        modules = get_modules([self.modules_folder])
        vivado_ip_cores = VivadoIpCores(modules, self.project_folder)

        assert vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()

    @mock.patch("tsfpga.vivado_ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_ip_core_file_is_removed(self, create):
        delete(self.hest_tcl)
        modules = get_modules([self.modules_folder])
        vivado_ip_cores = VivadoIpCores(modules, self.project_folder)

        assert vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()

    @mock.patch("tsfpga.vivado_ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_ip_core_file_is_changed(self, create):
        create_file(self.apa_tcl, "blaha blaha")
        modules = get_modules([self.modules_folder])
        vivado_ip_cores = VivadoIpCores(modules, self.project_folder)

        assert vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import copy
from glob import glob
from os.path import basename, isfile, join, exists, isdir

from tsfpga.constraint import Constraint
from tsfpga.hdl_file import HdlFile
from tsfpga.registers import from_json
from tsfpga.system_utils import load_python_module


class BaseModule:
    """
    Base class for handling a HDL module with RTL code, constraints, etc.

    Files are gathered from a lot of different subfolders, to accomodate for projects having different catalog structure.
    """

    def __init__(self, path, library_name_has_lib_suffix=False, default_registers=None):
        """
        :param path: Path to the module folder.
        :param library_name_has_lib_suffix: If set, the library name will be <module name>_lib,
                                            otherwise it is just <module name>.
        :param default_registers: A dictionary of Register objects.
        """
        self.path = path
        self.name = basename(self.path)
        self.library_name = self._get_library_name(self.name, library_name_has_lib_suffix)

        # Note: Likely mutable object, need to create deep copy before using.
        self._default_registers = default_registers
        self._registers = None

    @staticmethod
    def _get_file_list(folders, file_endings):
        """
        Returns a list of files given a list of folders.
        """
        files = []
        for folder in folders:
            for filename in glob(join(folder, "*")):
                if isfile(filename) and filename.lower().endswith(file_endings):
                    files.append(filename)
        return files

    def _get_hdl_file_list(self, folders):
        """
        Return a list of HDL file objects.
        """
        return [HdlFile(filename) for filename in self._get_file_list(folders, HdlFile.file_endings)]

    @property
    def registers(self):
        if self._registers is not None:
            # Only create object once
            return self._registers

        json_file = join(self.path, self.name + "_regs.json")
        if exists(json_file):
            self._registers = from_json(self.name, json_file, copy.deepcopy(self._default_registers))
            return self._registers

    def create_regs_vhdl_package(self):
        """
        Create a VHDL package in this module with register definitions.
        """
        if self.registers is not None:
            self.registers.create_vhdl_package(self.path)

    def get_synthesis_files(self):
        """
        List of files that should be included in a synthesis project.
        """
        self.create_regs_vhdl_package()

        folders = [
            self.path,
            join(self.path, "src"),
            join(self.path, "rtl"),
            join(self.path, "hdl", "rtl"),
            join(self.path, "hdl", "package"),
        ]
        return self._get_hdl_file_list(folders)

    def get_simulation_files(self, include_tests=True):
        """
        List of files that should be included in a simulation project.

        When include_tests is False the test folder is not included.
        The use case of include_tests is when testing a primary module
        that depends on other secondary modules we may want to compile
        the simulation files of the secondary modules but not their
        test files.

        Note: test-files are considered private to the module and
        should never be used by other modules.
        """
        self.create_regs_vhdl_package()

        test_folders = [
            join(self.path, "sim"),
        ]

        if include_tests:
            test_folders += [join(self.path, "rtl", "tb"),
                             join(self.path, "test")]

        return self.get_synthesis_files() + self._get_hdl_file_list(test_folders)

    @staticmethod
    def _get_library_name(module_name, library_name_has_lib_suffix):
        """
        By praxis VHDL libraries should be named e.g. ieee and not ieee_lib.
        However using <module_name>_lib is very common.
        """
        if library_name_has_lib_suffix:
            return module_name + "_lib"
        return module_name

    def setup_simulations(self, vunit_proj, **kwargs):
        """
        Setup local configuration of this module's test benches.
        Should be overriden by modules that have any test benches that operate via generics.
        """

    def get_ip_core_files(self):
        """
        Get a list of TCL files that set up the IP cores from this module.
        """
        folders = [
            join(self.path, "ip_cores"),
        ]
        file_endings = ("tcl")
        return self._get_file_list(folders, file_endings)

    def get_scoped_constraints(self):
        """
        Get a list of constraints that will be applied to a certain entity within the module.
        """
        scoped_constraints_folders = [
            join(self.path, "scoped_constraints"),
            join(self.path, "entity_constraints"),
            join(self.path, "hdl", "constraints"),
        ]
        constraints_file_endings = ("tcl", "xdc")
        constraint_files = self._get_file_list(scoped_constraints_folders, constraints_file_endings)

        constraints = []
        if constraint_files:
            synthesis_files = self.get_synthesis_files()
            for constraint_file in constraint_files:
                # Scoped constraints often depend on clocks having been created by another constraint
                # file before they can work. Set processing order to "late" to make this more probable.
                constraint = Constraint(constraint_file, scoped_constraint=True, processing_order="late")
                constraint.validate_scoped_entity(synthesis_files)
                constraints.append(constraint)
        return constraints

    def __str__(self):
        return self.name + ": " + self.path


def iterate_module_folders(modules_folders):
    for modules_folder in modules_folders:
        for module_folder in glob(join(modules_folder, "*")):
            if isdir(module_folder):
                yield module_folder


def get_module_object(path, name, library_name_has_lib_suffix, default_registers):
    module_file = join(path, "module_" + name + ".py")

    if exists(module_file):
        return load_python_module(module_file).Module(path,
                                                      library_name_has_lib_suffix,
                                                      default_registers)
    return BaseModule(path, library_name_has_lib_suffix, default_registers)


def get_modules(modules_folders,
                names_include=None,
                names_avoid=None,
                library_name_has_lib_suffix=False,
                default_registers=None):
    """
    Get a list of Module objects based on the source code folders.

    :param modules_folders: A list of paths where your modules are located.
    :param names_include: If specified, only modules with these names will be included.
    :param names_avoid: If specified, modules with these names will be discarded.
    :param library_name_has_lib_suffix: See BaseModule.
    :param default_registers: See BaseModule.
    """
    modules = []

    for module_folder in iterate_module_folders(modules_folders):
        module_name = basename(module_folder)
        if (names_include is None or module_name in names_include) \
                and (names_avoid is None or module_name not in names_avoid):
            modules.append(get_module_object(module_folder,
                                             module_name,
                                             library_name_has_lib_suffix,
                                             default_registers))

    return modules



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import join, dirname
import pytest
import unittest

from tsfpga.git_utils import find_git_files
from tsfpga.system_utils import create_file, delete


THIS_DIR = dirname(__file__)


def open_file_with_encoding(file):
    print(file)
    with open(file, "r", encoding="ascii") as file_handle:
        file_handle.read()


def test_all_checked_in_files_are_properly_encoded():
    """
    To avoid problems with different editors and different file encodings, all checked in files should
    contain only ASCII characters.
    """
    for file in find_git_files():
        open_file_with_encoding(file)


def check_file_for_tab_character(file):
    test_ok = True
    with open(file) as file_handle:
        for idx, line in enumerate(file_handle.readlines()):
            if "\t" in line:
                test_ok = False
                print("TAB charatcher (\\t) on line %i in %s" % (idx + 1, file))
    return test_ok


def test_no_checked_in_files_contain_tabs():
    """
    To avoid problems with files looking different in different editors, no checked in files may
    contain TAB characters.
    """
    test_ok = True
    for file in find_git_files():
        test_ok &= check_file_for_tab_character(file)
    assert test_ok


def check_file_for_carriage_return(file):
    test_ok = True
    with open(file, newline="") as file_handle:
        if "\r" in file_handle.read():
            test_ok = False
            print("Windows style line breaks (\\r\\n aka CR/LF) in " + file)
    return test_ok


def test_no_checked_in_files_contain_carriage_return():
    """
    All checked in files should use UNIX style line breaks (\n not \r\n).
    Some Linux editors and tools will display or interpret the \r as something other than a line break.
    """
    test_ok = True
    for file in find_git_files():
        test_ok &= check_file_for_carriage_return(file)
    assert test_ok


def check_file_for_trailing_whitespace(file):
    test_ok = True
    with open(file) as file_handle:
        for idx, line in enumerate(file_handle.readlines()):
            if " \n" in line:
                test_ok = False
                print("Trailing whitespace on line %i in %s" % (idx + 1, file))
    return test_ok


def test_no_checked_in_files_contain_trailing_whitespace():
    """
    Trailing whitespace is not allowed.
    Some motivation here: https://softwareengineering.stackexchange.com/questions/121555/why-is-trailing-whitespace-a-big-deal
    """
    test_ok = True
    for file in find_git_files():
        test_ok &= check_file_for_trailing_whitespace(file)
    assert test_ok


class TestFileFormat(unittest.TestCase):

    file = join(THIS_DIR, "temp_file_for_test.txt")

    def setUp(self):
        delete(self.file)

    def test_open_file_with_encoding_should_raise_exception_on_bad_file(self):
        """
        Sanity check that the function we use actually triggers on bad files.
        """
        with open(self.file, "w", encoding="utf-8") as file_handle:
            data = "\N{LATIN CAPITAL LETTER O WITH DIAERESIS}"  # Swedish word for island = non-ASCII character
            file_handle.write(data)

        with pytest.raises(UnicodeDecodeError):
            open_file_with_encoding(self.file)

    def test_check_file_for_tab_character_should_fail_on_bad_file(self):
        """
        Sanity check that the function we use actually triggers on bad files.
        """
        data = "Apa\thest"
        create_file(self.file, data)
        assert not check_file_for_tab_character(self.file)

    def test_check_file_for_carriage_return_should_fail_on_bad_file(self):
        """
        Sanity check that the function we use actually triggers on bad files.
        """
        data = b"Apa\r\nhest"
        with open(self.file, "wb") as file_handle:
            file_handle.write(data)
        assert not check_file_for_carriage_return(self.file)

    def test_check_file_for_trailing_whitespace(self):
        """
        Sanity check that the function we use actually triggers on bad files.
        """
        data = "Apa \nhest    \nzebra"
        create_file(self.file, data)
        assert not check_file_for_trailing_whitespace(self.file)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

"""
A set of reusable functions for working with HDL projects.
"""


from os.path import dirname, abspath, join
from tsfpga.about import get_version


THIS_DIR = dirname(__file__)
ROOT = abspath(join(THIS_DIR, ".."))  # Root of the git repo

TSFPGA_MODULES = join(ROOT, "modules")
TSFPGA_TCL = join(THIS_DIR, "tcl")

TSFPGA_EXAMPLES = join(ROOT, "examples")
TSFPGA_EXAMPLE_MODULES = join(TSFPGA_EXAMPLES, "modules")
TSFPGA_EXAMPLE_MODULES_WITH_IP = join(TSFPGA_EXAMPLES, "modules_with_ip")

ALL_TSFPGA_MODULES_FOLDERS = [TSFPGA_MODULES, TSFPGA_EXAMPLE_MODULES, TSFPGA_EXAMPLE_MODULES_WITH_IP]

__version__ = get_version()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


class RegisterVhdlGenerator:

    def __init__(self, register_list):
        self.register_list = register_list

    @staticmethod
    def _comment(comment):
        return f"-- {comment}\n"

    def _header(self):
        vhdl = self._comment(self.register_list.generated_info())
        return vhdl

    def _register_constant_name(self, register):
        return self.register_list.name + "_" + register.name

    def _register_indexes(self):
        vhdl = ""
        for register in self.register_list.iterate_registers():
            constant = self._register_constant_name(register)
            vhdl += f"  constant {constant} : integer := {register.idx};\n"

        return vhdl

    def _register_map(self):
        num_regs = str(len(self.register_list.registers))
        map_name = self.register_list.name + "_reg_map"

        register_definitions = []
        for register in self.register_list.iterate_registers():
            constant = self._register_constant_name(register)
            mode = register.mode
            register_definitions.append(f"  (idx => {constant}, reg_type => {mode})")

        vhdl = f"  constant {map_name} : reg_definition_vec_t(0 to {num_regs} - 1) := (\n  "
        vhdl += ",\n  ".join(register_definitions)
        vhdl += "\n  );\n"
        vhdl += "\n"
        vhdl += f"  subtype {self.register_list.name}_regs_t is reg_vec_t({map_name}'range);\n"
        vhdl += f"  constant {self.register_list.name}_regs_zero : \
{self.register_list.name}_regs_t := (others => (others => '0'));\n"
        vhdl += f"  subtype {self.register_list.name}_reg_was_written_t is std_logic_vector({map_name}'range);\n"
        return vhdl

    def _register_bits(self):
        vhdl = ""
        for register in self.register_list.iterate_registers():
            for bit in register.bits:
                name = self._register_constant_name(register) + "_" + bit.name
                vhdl += f"  constant {name} : integer := {bit.idx};\n"
            if register.bits:
                vhdl += "\n"

        return vhdl

    def get_package(self):
        pkg_name = self.register_list.name + "_regs_pkg"

        vhdl = f"""
{self._header()}

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library reg_file;
use reg_file.reg_file_pkg.all;


package {pkg_name} is

{self._register_indexes()}
{self._register_map()}
{self._register_bits()}
end;"""

        return vhdl



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import datetime
from collections import OrderedDict

from tsfpga.git_utils import git_commands_are_available, get_git_commit
from tsfpga.svn_utils import svn_commands_are_available, get_svn_revision_information


class Bit:

    def __init__(self, idx, name, description):
        self.idx = idx
        self.name = name
        self.description = description


# mode: (mode_readable, description)
REGISTER_MODES = {
    "r": ("Read", "PS can read a value that PL provides."),
    "w": ("Write", "PS can write a value that is available for PL usage."),
    "r_w": (
        "Read, Write",
        "PS can write a value and read it back. The written value is available for PL usage."),
    "wpulse": ("Write-pulse", "PS can write a value that is asserted for one clock cycle in PL."),
    "r_wpulse": (
        "Read, Write-pulse",
        "PS can read a value that PL provides. "
        "PS can write a value that is asserted for one clock cycle in PL."),
}


class Register:

    def __init__(self, name, idx, mode, description=""):
        self.name = name
        self.idx = idx
        self.mode = mode
        self.description = description
        self.bits = []

    def append_bit(self, bit_name, bit_description):
        idx = len(self.bits)
        bit = Bit(idx, bit_name, bit_description)

        self.bits.append(bit)
        return bit

    @property
    def mode_readable(self):
        """
        A human-readable representation of the register mode.
        """
        return REGISTER_MODES[self.mode][0]

    @property
    def address(self):
        address_int = 4 * self.idx
        num_nibbles_needed = 4
        formatting_string = "0x{:0%iX}" % num_nibbles_needed
        return formatting_string.format(address_int)

    @property
    def is_ps_readable(self):
        """
        Return true if the register is readable by PS.
        """
        return self.mode in ["r", "r_w", "r_wpulse"]

    @property
    def is_ps_writeable(self):
        """
        Return true if the register is writeable by PS.
        """
        return self.mode in ["w", "r_w", "wpulse", "r_wpulse"]


class RegisterList:

    def __init__(self, name):
        self.name = name
        self.registers = OrderedDict()

    def append(self, register_name, mode):
        idx = len(self.registers)
        register = Register(register_name, idx, mode)

        self.registers[register_name] = register
        return register

    @staticmethod
    def generated_info():
        """
        A string informing the user that a file is automatically generated
        """
        info = "File automatically generated by tsfpga"
        return info

    @staticmethod
    def generated_source_info():
        """
        A string containing info about the source of the generated register information
        """
        commit_info = ""
        if git_commands_are_available():
            commit_info = "from " + get_git_commit() + " "
        elif svn_commands_are_available():
            commit_info = "from " + get_svn_revision_information() + " "

        time_info = datetime.datetime.now().strftime("%Y-%m-%d %H:%M")

        info = f"Generated by tsfpga {commit_info}at {time_info}."
        return info

    def iterate_registers(self):
        for register in self.registers.values():
            yield register



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import dirname, join, exists
import pytest
from subprocess import CalledProcessError
import sys
import unittest

import tsfpga
from tsfpga.constraint import Constraint
from tsfpga.module import get_modules
from tsfpga.system_utils import create_file, delete, run_command
from tsfpga.test import file_contains_string
from tsfpga.vivado_project import VivadoProject


THIS_DIR = dirname(__file__)


def test_building_artyz7_project(tmpdir):
    build_py = join(tsfpga.TSFPGA_EXAMPLES, "build.py")
    cmd = [
        sys.executable,
        build_py,
        "artyz7",
        "--project-path", tmpdir,
        "--output-path", tmpdir
    ]
    run_command(cmd, cwd=tsfpga.ROOT)
    assert exists(join(tmpdir, "artyz7.bit"))
    assert exists(join(tmpdir, "artyz7.bin"))
    assert exists(join(tmpdir, "artyz7.hdf"))


class TestBasicProject(unittest.TestCase):

    part = "xc7z020clg400-1"
    modules_folder = join(THIS_DIR, "modules")
    project_folder = join(THIS_DIR, "vivado")

    top_file = join(modules_folder, "apa", "test_proj_top.vhd")
    top_template = """
library ieee;
use ieee.std_logic_1164.all;

library resync;


entity test_proj_top is
  port (
    clk_in : in std_logic;
    input : in std_logic;
    clk_out : in std_logic;
    output : out std_logic
  );
end entity;

architecture a of test_proj_top is
  signal input_p1 : std_logic;
begin

  pipe_input : process
  begin
    wait until rising_edge(clk_in);
    input_p1 <= input;
  end process;

{assign_output}

end architecture;
"""

    resync = """
  assign_output : entity resync.resync_level
  port map (
    data_in => input_p1,

    clk_out => clk_out,
    data_out => output
  );"""

    unhandled_clock_crossing = """
  assign_output : process
  begin
    wait until rising_edge(clk_out);
    output <= input_p1;
  end process;"""

    constraint_file = join(modules_folder, "apa", "test_proj_pinning.tcl")
    constraints = """
set_property -dict {package_pin H16 iostandard lvcmos33} [get_ports clk_in]
set_property -dict {package_pin P14 iostandard lvcmos33} [get_ports input]
set_property -dict {package_pin K17 iostandard lvcmos33} [get_ports clk_out]
set_property -dict {package_pin T16 iostandard lvcmos33} [get_ports output]

# 200 MHz
create_clock -period 5 -name clk_in [get_ports clk_in]
create_clock -period 5 -name clk_out [get_ports clk_out]
"""

    def setUp(self):
        delete(self.modules_folder)
        delete(self.project_folder)

        self.top = self.top_template.format(assign_output=self.resync)  # Default top level

        create_file(self.top_file, self.top)
        create_file(self.constraint_file, self.constraints)
        constraints = [Constraint(self.constraint_file)]

        self.modules = get_modules([self.modules_folder, tsfpga.TSFPGA_MODULES])
        self.proj = VivadoProject(name="test_proj", modules=self.modules, part=self.part, constraints=constraints)
        self.proj.create(self.project_folder)

        self.log_file = join(self.project_folder, "vivado.log")

    def test_create_project(self):
        pass

    def test_synth_project(self):
        self.proj.build(self.project_folder, synth_only=True)

    def test_synth_should_fail_if_source_code_does_not_compile(self):
        with open(self.top_file, "a") as file_handle:
            file_handle.write("garbage\napa\nhest")

        with pytest.raises(CalledProcessError):
            self.proj.build(self.project_folder, synth_only=True)
        assert file_contains_string(self.log_file, "\nERROR: Run synth_1 failed.")

    def test_synth_with_unhandled_clock_crossing_should_fail(self):
        top = self.top_template.format(assign_output=self.unhandled_clock_crossing)
        create_file(self.top_file, top)

        with pytest.raises(CalledProcessError):
            self.proj.build(self.project_folder, synth_only=True)
        assert file_contains_string(self.log_file, "\nERROR: Unhandled clock crossing in synth_1 run.")

    def test_build_project(self):
        self.proj.build(self.project_folder, self.project_folder)
        assert exists(join(self.project_folder, self.proj.name + ".bit"))
        assert exists(join(self.project_folder, self.proj.name + ".bin"))



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import join

from tsfpga.vivado_utils import to_tcl_path


class VivadoTcl:
    """
    Class with methods for translating a set of sources into Vivado TCL
    """

    def __init__(self, name,):
        self.name = name

    # pylint: disable=too-many-arguments
    def create(self,
               project_folder,
               modules,
               part,
               top,
               generics=None,
               constraints=None,
               tcl_sources=None,
               build_step_hooks=None,
               ip_cache_path=None):
        tcl = f"create_project {self.name} {to_tcl_path(project_folder)} -part {part}\n"
        tcl += "set_property target_language VHDL [current_project]\n"
        if ip_cache_path is not None:
            tcl += f"config_ip_cache -use_cache_location {to_tcl_path(ip_cache_path)}\n"
        # Default value for when opening project in GUI.
        # Will be overwritten if using build() function.
        tcl += "set_param general.maxThreads 4\n"
        tcl += "set_property STEPS.WRITE_BITSTREAM.ARGS.BIN_FILE true [get_runs impl_1]\n"
        tcl += "\n"
        tcl += self._add_modules(modules)
        tcl += "\n"
        tcl += self._add_generics(generics)
        tcl += "\n"
        all_constraints = self._iterate_constraints(modules, constraints)
        tcl += self._add_constraints(all_constraints)
        tcl += "\n"
        tcl += self._add_tcl_sources(tcl_sources)
        tcl += "\n"
        tcl += self._add_build_step_hooks(build_step_hooks)
        tcl += "\n"
        tcl += f"set_property top {top} [current_fileset]\n"
        tcl += "reorder_files -auto -disable_unused\n"
        tcl += "\n"
        tcl += "exit\n"
        return tcl

    def _add_modules(self, modules):
        tcl = ""
        for module in modules:
            vhdl_files = []
            verilog_source_files = []
            for hdl_file in module.get_synthesis_files():
                if hdl_file.is_vhdl:
                    vhdl_files.append(to_tcl_path(hdl_file.filename))
                elif hdl_file.is_verilog_source:
                    verilog_source_files.append(to_tcl_path(hdl_file.filename))
                else:
                    raise NotImplementedError("Can not handle file: " + hdl_file.filename)
                    # Verilog headers do not need to be handled at all if the
                    # source file that uses them is in the same directory. If
                    # it is not, the path needs to be added to include_dirs with
                    # a tcl command like:
                    #   set_property include_dirs {/some/path /some/other/path} [current_fileset]
                    # See https://www.xilinx.com/support/answers/54006.html

                    # Encrypted source files (verilog (.vp?), VHDL) I do not know how
                    # to handle, since I have no use case for it at the moment.

            if vhdl_files:
                tcl += "read_vhdl -library %s -vhdl2008 {%s}\n" % \
                    (module.library_name, " ".join(vhdl_files))
            if verilog_source_files:
                tcl += "read_verilog {%s}\n" % " ".join(verilog_source_files)

            tcl += self._add_tcl_sources(module.get_ip_core_files())
        return tcl

    @staticmethod
    def _add_tcl_sources(tcl_sources):
        if tcl_sources is None:
            return ""

        tcl = ""
        for tcl_source_file in tcl_sources:
            tcl += "source -notrace %s\n" % to_tcl_path(tcl_source_file)
        return tcl

    @staticmethod
    def _add_build_step_hooks(build_step_hooks):
        if build_step_hooks is None:
            return ""

        tcl = ""
        hook_steps_added = set()
        for build_step_hook in build_step_hooks:
            if build_step_hook.hook_step in hook_steps_added:
                message = f"Multiple TCL sources for hook step {build_step_hook.hook_step}: " + \
                    " ".join(map(str, build_step_hooks))
                raise ValueError(message)
            hook_steps_added.add(build_step_hook.hook_step)

            # Decoding the run name like this is not very nice for the case were the user has defined
            # many runs. But to solve that we would need to make structural changes, which I'm not sure about
            # at the moment. Will solve when there is an actual use case.
            run = "synth_1" if build_step_hook.step_is_synth else "impl_1"
            tcl += f"set_property {build_step_hook.hook_step} {to_tcl_path(build_step_hook.tcl_file)} [get_runs {run}]\n"
        return tcl

    @staticmethod
    def _add_generics(generics):
        """
        Generics are set accoring to this weird format: https://www.xilinx.com/support/answers/52217.html
        """
        if generics is None:
            return ""

        generic_list = []
        for name, value in generics.items():
            if isinstance(value, bool):
                generic_list.append("%s=%s" % (name, ("1'b1" if value else "1'b0")))
            else:
                generic_list.append("%s=%s" % (name, value))
        return "set_property generic {%s} [current_fileset]\n" % " ".join(generic_list)

    @staticmethod
    def _iterate_constraints(modules, constraints):
        for module in modules:
            for constraint in module.get_scoped_constraints():
                yield constraint

        if constraints is not None:
            for constraint in constraints:
                yield constraint

    @staticmethod
    def _add_constraints(constraints):
        tcl = ""
        for constraint in constraints:
            file = to_tcl_path(constraint.file)
            ref_flags = "" if constraint.ref is None else (f"-ref {constraint.ref} ")
            managed_flags = "" if file.endswith("xdc") else "-unmanaged "

            tcl += f"read_xdc {ref_flags}{managed_flags}{file}\n"
            tcl += f"set_property PROCESSING_ORDER {constraint.processing_order} [get_files {file}]\n"

            if constraint.used_in == "impl":
                tcl += f"set_property used_in_synthesis false [get_files {file}]\n"
            elif constraint.used_in == "synth":
                tcl += f"set_property used_in_implementation false [get_files {file}]\n"

        return tcl

    def build(self, project_file, output_path, num_threads, generics=None, synth_only=False):
        synth_run = "synth_1"
        impl_run = "impl_1"
        num_threads = min(num_threads, 8)  # Max value in Vivado 2017.4. set_param will give an error if higher number.

        tcl = f"open_project {to_tcl_path(project_file)}\n"
        tcl += f"set_param general.maxThreads {num_threads}\n"
        tcl += "\n"
        tcl += self._add_generics(generics)
        tcl += "\n"
        tcl += self._synthesis(synth_run, num_threads)
        tcl += "\n"
        if not synth_only:
            tcl += self._impl(impl_run, num_threads)
            tcl += "\n"
            tcl += self._bitstream(output_path)
            tcl += self._hwdef(output_path)
            tcl += "\n"
        tcl += "exit\n"
        return tcl

    @staticmethod
    def _run(run, num_threads):
        tcl = f"reset_run {run}\n"
        tcl += f"launch_runs {run} -jobs {num_threads}\n"
        tcl += "wait_on_run %s\n" % run
        tcl += "\n"
        tcl += "if {[get_property PROGRESS [get_runs %s]] != \"100%%\"} {\n" % run
        tcl += f"  puts \"ERROR: Run {run} failed.\"\n"
        tcl += "  exit 1\n"
        tcl += "}\n"
        tcl += "\n"
        tcl += f"open_run {run}\n"
        tcl += f"set run_directory [get_property DIRECTORY [get_runs {run}]]\n"
        return tcl

    def _check_clock_interaction(self, run):
        tcl = r"if {[regexp {\(unsafe\)} [report_clock_interaction -delay_type min_max -return_string]]} "
        tcl += "{\n"
        tcl += f"  puts \"ERROR: Unhandled clock crossing in {run} run. See reports in ${{run_directory}}\"\n"
        tcl += "\n"
        tcl += self._save_clock_interaction_report()
        tcl += self._save_timing_report()
        tcl += "\n"
        tcl += "  exit 1\n"
        tcl += "}\n"
        return tcl

    @staticmethod
    def _save_clock_interaction_report():
        tcl = "set output_file [file join ${run_directory} \"clock_interaction.rpt\"]\n"
        tcl += "report_clock_interaction -delay_type min_max -file ${output_file}\n"
        return tcl

    @staticmethod
    def _save_timing_report():
        tcl = "set output_file [file join ${run_directory} \"timing_summary.rpt\"]\n"
        tcl += "report_timing_summary -file ${output_file}\n"
        return tcl

    def _check_timing(self, run):
        tcl = "if {[expr {[get_property SLACK [get_timing_paths -delay_type min_max]] < 0}]} {\n"
        tcl += f"  puts \"ERROR: Timing not OK after {run} run. See reports in ${{run_directory}}\"\n"
        tcl += "\n"
        tcl += self._save_timing_report()
        tcl += "\n"
        tcl += "  exit 1\n"
        tcl += "}\n"
        return tcl

    @staticmethod
    def _report_utilization():
        tcl = "set output_file [file join ${run_directory} \"hierarchical_utilization.rpt\"]\n"
        tcl += "report_utilization -hierarchical -hierarchical_depth 4 -file ${output_file}\n"
        return tcl

    def _synthesis(self, run, num_threads):
        tcl = self._run(run, num_threads)
        tcl += "\n"
        tcl += self._report_utilization()
        tcl += "\n"
        tcl += self._check_clock_interaction(run)
        return tcl

    def _impl(self, run, num_threads):
        tcl = self._run(run, num_threads)
        tcl += "\n"
        tcl += self._report_utilization()
        tcl += "\n"
        tcl += self._check_timing(run)
        return tcl

    def _bitstream(self, output_path):
        bit_file = to_tcl_path(join(output_path, self.name))  # Vivado will append the appropriate file ending
        tcl = f"write_bitstream -force -bin {bit_file}\n"
        return tcl

    def _hwdef(self, output_path):
        # Vivado will append the wrong file ending (.hwdef) unless specified
        hwdef_file = to_tcl_path(join(output_path, self.name + ".hdf"))
        tcl = "write_hwdef -force %s\n" % hwdef_file
        return tcl



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import copy
from os.path import dirname, join
import unittest
import pytest

from tsfpga.system_utils import create_file
from tsfpga.registers import load_json_file, from_json, get_default_registers


THIS_DIR = dirname(__file__)


def test_deep_copy_of_register_actually_copies_everything():
    registers = get_default_registers()
    registers_copy = copy.deepcopy(registers)

    registers_copy["config"].description = "Dummy"
    registers_copy["config"].bits.append("dummy object")

    assert registers["config"].description == "Configuration register."
    assert len(registers["config"].bits) == 0


class TestRegisters(unittest.TestCase):

    module_name = "sensor"
    json_file = join(THIS_DIR, "sensor_regs.json")
    json_data = """\
{
  "conf": {
    "description": "Set configuration bits",
    "mode": "r_w",
    "bits": {
      "enable": "Enable things",
      "disable": ""
    }
  },
  "cmd": {
    "mode": "w"
  },
  "irq": {
    "description": "Interrupt register",
    "mode": "r_w",
    "bits": {
      "bad": "",
      "not_good": ""
    }
  }%s
}
"""

    def setUp(self):
        create_file(self.json_file, self.json_data % "")

    def test_order_of_registers_and_bits(self):
        json_registers = from_json(self.module_name, self.json_file)
        registers = list(json_registers.register_list.registers.values())

        assert registers[0].name == "conf"
        assert registers[1].name == "cmd"
        assert registers[2].name == "irq"

        assert registers[0].bits[0].name == "enable"
        assert registers[0].bits[1].name == "disable"
        assert registers[1].bits == []
        assert registers[2].bits[0].name == "bad"
        assert registers[2].bits[1].name == "not_good"

    def test_default_registers(self):
        default_registers = get_default_registers()
        num_default_registers = len(default_registers)
        json_registers = from_json(self.module_name, self.json_file, default_registers)

        # The registers from this test are appended at the end
        assert json_registers.register_list.registers["conf"].idx == num_default_registers
        assert json_registers.register_list.registers["cmd"].idx == num_default_registers + 1
        assert json_registers.register_list.registers["irq"].idx == num_default_registers + 2

    def test_load_nonexistent_json_file_should_raise_exception(self):
        filename = self.json_file + "apa"
        with pytest.raises(FileNotFoundError) as exception_info:
            load_json_file(self.json_file + "apa")
        assert str(exception_info.value) == f"Requested json file does not exist: {filename}"

    def test_load_dirty_json_file_should_raise_exception(self):
        data = self.json_data % "apa"
        create_file(self.json_file, data)

        with pytest.raises(ValueError) as exception_info:
            load_json_file(self.json_file)
        assert str(exception_info.value).startswith(f"Error while parsing JSON file {self.json_file}:\nExpecting ',' delimiter")

    def test_no_mode_field_should_raise_exception(self):
        extras = """,
  "apa": {
    "description": "w"
  }"""
        data = self.json_data % extras
        create_file(self.json_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_json(self.module_name, self.json_file)
        assert str(exception_info.value) == f"Register apa in {self.json_file} does not have mode field"

    def test_two_registers_with_same_name_should_raise_exception(self):
        extras = """,
  "conf": {
    "mode": "w"
  }"""
        data = self.json_data % extras
        create_file(self.json_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_json(self.module_name, self.json_file)
        assert str(exception_info.value) == f"Error while parsing JSON file {self.json_file}:\nDuplicate key conf"

    def test_two_bits_with_same_name_should_raise_exception(self):
        extras = """,
  "test_reg": {
    "mode": "w",
    "bits": {
      "test_bit": "Declaration 1",
      "test_bit": "Declaration 2"
    }
  }"""
        data = self.json_data % extras
        create_file(self.json_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_json(self.module_name, self.json_file)
        assert str(exception_info.value) == f"Error while parsing JSON file {self.json_file}:\nDuplicate key test_bit"

    def test_overriding_default_register(self):
        extras = """,
  "config": {
    "description": "apa"
  }"""
        data = self.json_data % extras
        create_file(self.json_file, data)
        json_registers = from_json(self.module_name, self.json_file, get_default_registers())

        assert json_registers.register_list.registers["config"].description == "apa"

    def test_changing_mode_of_default_register_should_raise_exception(self):
        extras = """,
  "config": {
    "mode": "w"
  }"""
        data = self.json_data % extras
        create_file(self.json_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_json(self.module_name, self.json_file, get_default_registers())
        assert str(exception_info.value) == f"Overloading register config in {self.json_file}, one can not change mode from default"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import unittest

from tsfpga.register_c_generator import RegisterCGenerator
from tsfpga.register_cpp_generator import RegisterCppGenerator
from tsfpga.register_html_generator import RegisterHtmlGenerator
from tsfpga.register_list import RegisterList
from tsfpga.register_vhdl_generator import RegisterVhdlGenerator


class TestRegisterCodeGeneration(unittest.TestCase):

    def setUp(self):
        self.register_list = RegisterList("sensor")

        register = self.register_list.append("conf", "r_w")
        register.description = "conf desc"
        register.append_bit("bit_0", "conf bit 0 desc")
        register.append_bit("bit_1", "")

        register = self.register_list.append("addr", "w")
        register.description = "addr desc is _very_ good"

    def test_generated_html_contains_all_fields_in_correct_order(self):
        expected = """
  <tr>
    <td><strong>conf</strong></td>
    <td>0x0000</td>
    <td>Read, Write</td>
    <td>conf desc</td>
  </tr>
  <tr>
    <td>&nbsp;&nbsp;<em>bit_0</em></td>
    <td>0</td>
    <td></td>
    <td>conf bit 0 desc</td>
  </tr>
  <tr>
    <td>&nbsp;&nbsp;<em>bit_1</em></td>
    <td>1</td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td><strong>addr</strong></td>
    <td>0x0004</td>
    <td>Write</td>
    <td>addr desc is <em>very</em> good</td>
  </tr>
"""
        html_generator = RegisterHtmlGenerator(self.register_list)
        assert expected in html_generator.get_table()
        assert expected in html_generator.get_page()

    def test_generated_vhdl_contains_all_fields_in_correct_order(self):
        vhdl_generator = RegisterVhdlGenerator(self.register_list)
        expected = """
  constant sensor_conf : integer := 0;
  constant sensor_addr : integer := 1;
"""
        assert expected in vhdl_generator.get_package()
        expected = """
    (idx => sensor_conf, reg_type => r_w),
    (idx => sensor_addr, reg_type => w)
"""
        assert expected in vhdl_generator.get_package()
        expected = """
  constant sensor_conf_bit_0 : integer := 0;
  constant sensor_conf_bit_1 : integer := 1;
"""
        assert expected in vhdl_generator.get_package()

    def test_generated_c_contains_all_fields_in_correct_order(self):
        c_generator = RegisterCGenerator(self.register_list)
        expected = """
struct sensor_regs_t {
  uint32_t conf;
  uint32_t addr;
};
"""
        assert expected in c_generator.get_header()

        expected = """
#define SENSOR_CONF_BIT_0_BIT (0uL)
#define SENSOR_CONF_BIT_0     (1uL)

#define SENSOR_CONF_BIT_1_BIT (1uL)
#define SENSOR_CONF_BIT_1     (2uL)
"""
        assert expected in c_generator.get_header()

    def test_generated_cpp_contains_all_fields_in_correct_order(self):
        cpp_generator = RegisterCppGenerator(self.register_list)
        expected = """
  static const uint32_t conf_bit_0 = 1uL;
  static const uint32_t conf_bit_1 = 2uL;
"""
        assert expected in cpp_generator.get_interface()

        expected = """
uint32_t Sensor::get_conf() const
{
  return m_registers[0];
}

void Sensor::set_conf(uint32_t value) const
{
  m_registers[0] = value;
}

void Sensor::set_addr(uint32_t value) const
{
  m_registers[1] = value;
}
"""
        assert expected in cpp_generator.get_implementation()

        # Write-only register should not have getter method
        assert "get_addr" not in cpp_generator.get_implementation()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import abspath, dirname, exists, join
import toml


def create_configuration(output_path,
                         modules=None,
                         vunit_proj=None,
                         vivado_location=None,
                         ip_core_vivado_project_sources_directory=None):
    """
    Create a configuration file (vhdl_ls.toml) for the rust_hdl VHDL Language Server.

    Can be used with modules and an "empty" VUnit project, or with a complete VUnit
    project with all user files added.

    Execution of this function takes roughly 12 ms for a large project (62 modules and a
    VUnit project).

    :param output_path: Output folder.
    :param modules: A list of Module objects.
    :param vunit_proj: A VUnit project.
    :param vunit_proj: Vivado binary path. Will add unisim from this Vivado installation.
    :param ip_core_vivado_project_sources_directory: Path to the sources directory of a
        Vivado project that contains generated "simulation" and "synthesis" files
        of IP cores (the "generate_target" TCL command). See simulate.py for an example
        of using this.
    """
    toml_data = dict(libraries=dict())

    if modules is not None:
        for module in modules:
            vhd_file_wildcard = join(abspath(module.path), "**", "*.vhd")
            toml_data["libraries"][module.library_name] = dict(
                files=[vhd_file_wildcard])

    if vunit_proj is not None:
        for source_file in vunit_proj.get_compile_order():
            if source_file.library.name not in toml_data["libraries"]:
                toml_data["libraries"][source_file.library.name] = dict(
                    files=[])
            toml_data["libraries"][source_file.library.name]["files"].append(
                abspath(source_file.name))

    if vivado_location is not None:
        vcomponents_package = abspath(join(dirname(vivado_location),
                                           "..",
                                           "data",
                                           "vhdl",
                                           "src",
                                           "unisims",
                                           "unisim_retarget_VCOMP.vhd"))
        if not exists(vcomponents_package):
            raise FileNotFoundError("Could not find unisim file: " + vcomponents_package)

        toml_data["libraries"]["unisim"] = dict(files=[vcomponents_package])

    if ip_core_vivado_project_sources_directory is not None:
        ip_sources_dir = join(ip_core_vivado_project_sources_directory, "ip")
        if not exists(ip_sources_dir):
            raise FileNotFoundError("Could not find IP sources dir: " + ip_sources_dir)

        # Add file from the "synth" folder rather than "sim". It seems that "synth"
        # always contains a VHDL file while "sim" sometimes contains a Verilog file.
        vhd_file_wildcard = join(abspath(ip_sources_dir), "*", "synth", "*.vhd")
        toml_data["libraries"]["xil_defaultlib"] = dict(files=[vhd_file_wildcard])

    with open(join(output_path, "vhdl_ls.toml"), "w") as output_file_handle:
        toml.dump(toml_data, output_file_handle)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import dirname, join, abspath
import unittest

from tsfpga.system_utils import create_file, delete, run_command
from tsfpga.registers import from_json


THIS_DIR = dirname(__file__)


class TestRegisterCompilation(unittest.TestCase):
    """
    Functional test: JSON -> registers -> Code generation -> compilation
    """

    working_dir = abspath(join(THIS_DIR, "compilation"))
    include_dir = join(working_dir, "include")

    def setUp(self):
        delete(self.working_dir)
        module_name = "sensor"
        json_file = join(self.working_dir, "sensor_regs.json")
        json_data = """\
{
  "config": {
    "mode": "r_w",
    "bits": {
      "enable": "",
      "disable": ""
    }
  },
  "irq_status": {
    "mode": "r_w",
    "bits": {
      "bad": "",
      "not_good": ""
    }
  }
}
"""

        create_file(json_file, json_data)
        self.registers = from_json(module_name, json_file)

    def test_compiling_c_header(self):
        main_file = join(self.working_dir, "main.c")
        main = """\
#include "sensor_regs.h"

int main()
{
  struct sensor_regs_t regs;
  regs.config = SENSOR_CONFIG_ENABLE;
  regs.irq_status = (1 << SENSOR_IRQ_STATUS_BAD_BIT);
  return 0;
}
"""
        create_file(main_file, main)
        self.registers.create_c_header(self.include_dir)

        executable = join(self.working_dir, "sensor.o")
        cmd = [
            "gcc",
            main_file,
            "-o" + executable,
            "-I" + self.include_dir
        ]
        run_command(cmd)
        run_command([executable])

    def test_compiling_cpp(self):
        main_file = join(self.working_dir, "main.cpp")
        main = """\
#include <assert.h>

#include "include/sensor.h"

int main()
{
  long long data;
  volatile uint8_t *base_address = reinterpret_cast<volatile uint8_t *>(&data);
  fpga_regs::Sensor sensor = fpga_regs::Sensor(base_address);

  sensor.set_config(1337);
  sensor.set_irq_status(666);
  assert(sensor.get_config() == 1337);
  assert(sensor.get_irq_status() == 666);

  return 0;
}
"""
        create_file(main_file, main)
        self.registers.create_cpp_interface(self.include_dir)
        self.registers.create_cpp_header(self.include_dir)
        self.registers.create_cpp_implementation(self.working_dir)
        cpp_class_file = join(self.working_dir, "sensor.cpp")

        executable = join(self.working_dir, "sensor.o")
        cmd = [
            "g++",
            main_file,
            cpp_class_file,
            "-o" + executable,
            "-I" + self.working_dir
        ]
        run_command(cmd)
        run_command([executable])



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import dirname, abspath
import subprocess


def run_vivado_tcl(vivado_path, tcl_file, no_log_file=False):
    """
    Setting cwd ensures that any .log or .jou files produced are placed in
    the same directory as the TCL file that produced them.

    Subprocess has to be run with shell=True on Windows where vivado is a bat file.
    """
    cmd = "%s -mode batch -notrace -source %s" % (vivado_path, tcl_file)
    if no_log_file:
        cmd += " -nojournal -nolog"
    cwd = dirname(tcl_file)
    subprocess.check_call(cmd, cwd=cwd, shell=True)


def to_tcl_path(path):
    return abspath(path).replace("\\", "/")



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import hashlib
from os.path import abspath, exists, basename, join, dirname
from platform import system
from shutil import which

from tsfpga.vivado_utils import to_tcl_path, run_vivado_tcl
from tsfpga.system_utils import create_file, delete


class VivadoSimlib:

    _libraries = ["unisim", "secureip", "unimacro", "unifast", "xpm"]

    _tcl = "set_param general.maxthreads 4\n" \
        "compile_simlib -simulator {simulator_name} -simulator_exec_path {{{simulator_folder}}} " \
        "-family all -language all -library all -no_ip_compile -dir {{{output_path}}} -force"

    def __init__(self, vunit_proj, output_path, vivado_path="vivado"):
        self._vunit_proj = vunit_proj
        self._vivado_path = vivado_path

        self._find_simulator()
        self.output_path = join(output_path, self._get_version_tag())
        self._done_token = join(self.output_path, "done.txt")

    def compile_if_needed(self):
        if exists(self._done_token):
            return False
        self.compile()
        return True

    def compile(self):
        delete(self._done_token)
        print(f"Compiling Vivado simlib in {self.output_path}")
        tcl_file = join(self.output_path, "compile_simlib.tcl")
        tcl = self._tcl.format(simulator_name=self._simulator_name,
                               simulator_folder=to_tcl_path(self._simulator_folder),
                               output_path=to_tcl_path(self.output_path))
        create_file(tcl_file, tcl)
        run_vivado_tcl(self._vivado_path, tcl_file)
        create_file(self._done_token, "Done!")

    def add_to_vunit_project(self):
        for library in self._libraries:
            library_path = join(self.output_path, library)
            assert exists(library_path)
            self._vunit_proj.add_external_library(library, library_path)

    def _get_version_tag(self):
        """
        Return a version string on the format vivado-simlib-WW.XX.YY.ZZ,
        suitable for storage and versioning in Artifactory.
        """
        tag = "vivado-simlib-"
        tag += system().lower()
        tag += "." + self._get_tcl_hash()
        tag += "." + self._get_vivado_version_tag()
        tag += "." + self._get_simulator_tag()
        return tag

    def _get_tcl_hash(self):
        """
        Return a (partial) hash of the TCL script that is used to compile simlib.
        """
        return hashlib.md5(self._tcl.encode()).hexdigest()[0:8]

    def _get_vivado_version_tag(self):
        """
        Return e.g. vivado_2019_1
        """
        vivado_path = self._vivado_path
        if vivado_path == "vivado":
            vivado_path = which(vivado_path)
            assert vivado_path is not None, "Could not find vivado location"
        vivado_version = basename(abspath(join(dirname(vivado_path), "..")))
        return "vivado_" + vivado_version.replace(".", "_")

    def _find_simulator(self):
        # Update naming for Vivado's convetion
        vunit_simulator_name = self._vunit_proj.get_simulator_name()
        self._simulator_name = "riviera" if vunit_simulator_name == "rivierapro" else vunit_simulator_name

        # pylint: disable=protected-access
        self._simulator_folder = self._vunit_proj._simulator_class.find_prefix()

    def _get_simulator_tag(self):
        """
        Return e.g. modelsim_modeltech_pe_10_6c or riviera_riviera_pro_2018_10_x64.
        """
        simulator_version = basename(abspath(join(self._simulator_folder, "..")))
        return self._simulator_name + "_" + simulator_version.replace(".", "_").replace("-", "_").lower()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import join, dirname
import pytest
import re
import unittest

from tsfpga import TSFPGA_EXAMPLES

from tsfpga.system_utils import create_file, read_file, delete
from tsfpga.git_utils import find_git_files


THIS_DIR = dirname(__file__)


class CopyrightHeader:

    def __init__(self, copyright_holder, filename):
        self._copyright_holder = copyright_holder
        self._filename = filename

    def check_file(self):
        """
        Copyright comments should be correct. It should be followed by a blank line or another comment.
        """
        copyright_header_re = self.get_expected_copyright_header().replace("(", "\\(").replace(")", "\\)")
        regexp = re.compile(copyright_header_re + rf"($|\n|{self._comment_character})")
        with open(self._filename) as file_handle:
            data = file_handle.read()
        return regexp.match(data) is not None

    def get_expected_copyright_header(self):
        separator_line = self._comment_character + " " + "-" * (79 - len(self._comment_character))
        header = separator_line + "\n" + \
            self._comment_character + " " + f"Copyright (c) {self._copyright_holder}. All rights reserved." + "\n" + \
            separator_line + "\n"
        return header

    def fix_file(self):
        if self._is_suitable_for_insertion():
            self._insert_copyright_header()
        else:
            raise ValueError(f"Can not fix copyright header in file {self._filename}")

    @property
    def _comment_character(self):
        if self._filename.endswith(".py"):
            return "#"
        if self._filename.endswith(".vhd"):
            return "--"
        if self._filename.endswith(".tcl"):
            return "#"
        raise RuntimeError(f"Could not decide file ending of {self._filename}")

    def _is_suitable_for_insertion(self):
        """
        If the file does not begin with a comment, we consired it suitable to insert a new copyright header comment.
        """
        with open(self._filename) as file_handle:
            return not file_handle.readline().startswith(self._comment_character)

    def _insert_copyright_header(self):
        with open(self._filename) as file_handle:
            data = file_handle.read()
        data = self.get_expected_copyright_header() + "\n" + data
        with open(self._filename, "w") as file_handle:
            file_handle.write(data)


def files_to_check_for_copyright_header():
    file_endings = [".py", ".vhd", ".tcl"]
    exclude_directories = [join(TSFPGA_EXAMPLES, "modules", "artyz7", "tcl")]
    for file_ending in file_endings:
        for filename in find_git_files(file_ending=file_ending, exclude_directories=exclude_directories):
            yield filename


def test_copyright_header_of_all_checked_in_files():
    test_ok = True
    for filename in files_to_check_for_copyright_header():
        copyright_header_checker = CopyrightHeader("Lukas Vik", filename)
        if not copyright_header_checker.check_file():
            test_ok = False
            expected = copyright_header_checker.get_expected_copyright_header()
            print(f"Fail for {filename}.\nExpected:\n{expected}")
    assert test_ok


class TestCopyright(unittest.TestCase):

    file = join(THIS_DIR, "temp_file_for_test.vhd")

    def setUp(self):
        delete(self.file)

    def test_check_file(self):
        header = "-- " + "-" * 77 + "\n"
        header += "-- Copyright (c) Apa. All rights reserved.\n"
        header += "-- " + "-" * 77 + "\n"

        data = header
        create_file(self.file, data)
        copyright_header = CopyrightHeader("Apa", self.file)
        assert copyright_header.check_file()

        data = header + "non-comment on line after"
        create_file(self.file, data)
        copyright_header = CopyrightHeader("Apa", self.file)
        assert not copyright_header.check_file()

        data = header + "\nEmpty line and then non-comment"
        create_file(self.file, data)
        copyright_header = CopyrightHeader("Apa", self.file)
        assert copyright_header.check_file()

        data = header + "-- Further comment\n"
        create_file(self.file, data)
        copyright_header = CopyrightHeader("Apa", self.file)
        assert copyright_header.check_file()

    def test_fix_file_comment_insertion(self):
        data = "Apa\n"
        create_file(self.file, data)

        copyright_header = CopyrightHeader("Hest Hestsson", self.file)
        copyright_header.fix_file()

        data = read_file(self.file).split("\n")
        assert data[0] == "-- " + "-" * 77
        assert data[1] == "-- Copyright (c) Hest Hestsson. All rights reserved."
        assert data[2] == "-- " + "-" * 77
        assert data[3] == ""
        assert data[4] == "Apa"

    def test_fix_file_should_not_run_on_dirty_file(self):
        data = "-- Custom comment line\n\nApa\n"
        create_file(self.file, data)
        copyright_header = CopyrightHeader("A", self.file)

        with pytest.raises(ValueError) as exception_info:
            copyright_header.fix_file()
        assert str(exception_info.value) == f"Can not fix copyright header in file {self.file}"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import dirname, join
from unittest import TestCase
from unittest.mock import MagicMock, patch

from tsfpga.system_utils import delete
from tsfpga.vivado_simlib import VivadoSimlib


THIS_DIR = dirname(__file__)


class TestVivadoSimlib(TestCase):

    def setUp(self):
        self.output_path = join(THIS_DIR, "simlib")
        delete(self.output_path)

        self.vunit_mock = MagicMock()
        #  pylint: disable=protected-access
        self.vunit_mock.get_simulator_name.return_value = "rivierapro"
        self.vunit_mock._simulator_class.find_prefix.return_value = \
            "/opt/Aldec/Riviera-PRO-2018.10-x64/bin"

        self.vivado_path = "/tools/xilinx/Vivado/2019.1/bin/vivado"

    def test_should_not_recompile(self):
        vivado_simlib = VivadoSimlib(self.vunit_mock, self.output_path, self.vivado_path)
        self.assert_should_compile(vivado_simlib)
        self.assert_should_not_compile(vivado_simlib)

    def test_new_simulator_version_should_cause_recompile(self):
        vivado_simlib = VivadoSimlib(self.vunit_mock, self.output_path, self.vivado_path)
        self.assert_should_compile(vivado_simlib)
        vivado_simlib = VivadoSimlib(self.vunit_mock, self.output_path, self.vivado_path)
        self.assert_should_not_compile(vivado_simlib)

        #  pylint: disable=protected-access
        self.vunit_mock._simulator_class.find_prefix.return_value = \
            "/opt/Aldec/Riviera-PRO-2017.10-x64/bin"
        vivado_simlib = VivadoSimlib(self.vunit_mock, self.output_path, self.vivado_path)
        self.assert_should_compile(vivado_simlib)
        self.assert_should_not_compile(vivado_simlib)

    def test_new_vivado_version_should_cause_recompile(self):
        vivado_simlib = VivadoSimlib(self.vunit_mock, self.output_path, self.vivado_path)
        self.assert_should_compile(vivado_simlib)
        vivado_simlib = VivadoSimlib(self.vunit_mock, self.output_path, self.vivado_path)
        self.assert_should_not_compile(vivado_simlib)

        vivado_path = "/opt/xilinx/Vivado/2018.0/bin/vivado"
        vivado_simlib = VivadoSimlib(self.vunit_mock, self.output_path, vivado_path)
        self.assert_should_compile(vivado_simlib)
        self.assert_should_not_compile(vivado_simlib)

    @staticmethod
    def assert_should_compile(vivado_simlib):
        with patch("tsfpga.vivado_simlib.run_vivado_tcl", autospec=True) as mock:
            vivado_simlib.compile_if_needed()
            mock.assert_called_once()

    @staticmethod
    def assert_should_not_compile(vivado_simlib):
        with patch("tsfpga.vivado_simlib.run_vivado_tcl", autospec=True) as mock:
            vivado_simlib.compile_if_needed()
            mock.assert_not_called()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.module import BaseModule


class Module(BaseModule):

    def setup_simulations(self, vunit_proj, **kwargs):
        for tb_name in ["tb_fifo", "tb_afifo"]:
            tb = vunit_proj.library(self.library_name).test_bench(tb_name)
            for width in [8, 24]:
                for depth in [16, 1024]:
                    name = "width_%i.depth_%i" % (width, depth)
                    tb.add_config(name=name, generics=dict(width=width, depth=depth))



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import dirname, join
import pytest
import unittest

from tsfpga.module import get_modules
from tsfpga.system_utils import create_file, create_directory, delete
from tsfpga.vivado_project import VivadoProject


THIS_DIR = dirname(__file__)


class TestBasicProject(unittest.TestCase):

    project_folder = join(THIS_DIR, "vivado")
    modules_folder = join(THIS_DIR, "modules")

    def setUp(self):
        delete(self.modules_folder)
        delete(self.project_folder)

        self.b_vhd = create_file(join(self.modules_folder, "apa", "b.vhd"))

        self.modules = get_modules([self.modules_folder])
        self.proj = VivadoProject(name="name", modules=self.modules, part="part")

    def test_create_should_raise_exception_if_project_path_already_exists(self):
        create_directory(self.project_folder)
        with pytest.raises(ValueError) as exception_info:
            self.proj.create(self.project_folder)
        assert str(exception_info.value).startswith("Folder already exists")

    def test_build_should_raise_exception_if_project_does_not_exists(self):
        with pytest.raises(ValueError) as exception_info:
            self.proj.build(self.project_folder, synth_only=True)
        assert str(exception_info.value).startswith("Project file does not exist")

    def test_build_with_impl_run_should_raise_exception_if_no_output_path_is_given(self):
        with pytest.raises(ValueError) as exception_info:
            self.proj.build(self.project_folder)
        assert str(exception_info.value).startswith("Must specify output_path")



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.module import BaseModule


class Module(BaseModule):

    def setup_simulations(self, vunit_proj, **kwargs):
        tb = vunit_proj.library(self.library_name).test_bench("tb_axi_pkg")
        for data_width in [32, 64, 128]:
            for id_width in [0, 32]:
                name = "data_width_%i.id_width_%s" % (data_width, id_width)
                tb.add_config(name=name, generics=dict(data_width=data_width, id_width=id_width))

        for tb_name in ["tb_axil_pkg", "tb_axi_to_axil", "tb_axi_to_axil_bus_error"]:
            tb = vunit_proj.library(self.library_name).test_bench(tb_name)
            for data_width in [32, 64]:
                name = "data_width_%i" % data_width
                tb.add_config(name=name, generics=dict(data_width=data_width))

        tb = vunit_proj.library(self.library_name).test_bench("tb_axil_cdc")
        tb.add_config(name="master_clk_fast", generics=dict(master_clk_fast=True))
        tb.add_config(name="slave_clk_fast", generics=dict(slave_clk_fast=True))
        tb.add_config(name="same_clocks")



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import dirname, exists, join, relpath
from setuptools import setup, find_packages
import subprocess
import sys

ROOT = join(dirname(__file__))
sys.path.append(ROOT)
from tsfpga.about import get_version
from tsfpga.git_utils import find_git_files, git_local_changes_are_present


def main():
    setup(
        name="tsfpga",
        version=get_release_version(),
        description="A project platform for modern FPGA development",
        long_description=get_readme_description(),
        license="BSD 3-Clause License",
        author="Lukas Vik",
        author_email="2767848-LukasVik@users.noreply.gitlab.com",
        url="https://gitlab.com/truestream/tsfpga",
        python_requires=">=3.6",
        install_requires=get_package_requirements(),
        packages=find_packages(),
        package_data={"tsfpga": get_package_data()},
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "License :: OSI Approved :: BSD License",
            "Natural Language :: English",
            "Intended Audience :: Developers",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX :: Linux",
            "Topic :: Software Development :: Testing",
            "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        ],
        zip_safe=False
    )


def get_release_version():
    python_package_version = get_version()
    verify_release_version(python_package_version)
    return python_package_version


def verify_release_version(version):
    if git_local_changes_are_present():
        sys.exit("Must make release from clean repo")

    release_notes_file = join(ROOT, "doc", "release_notes", version + ".md")
    if not exists(release_notes_file):
        sys.exit("Could not find release notes file: " + release_notes_file)

    git_tag = "v" + version
    git_tags = subprocess.check_output(
        ["git", "tag", "--list", "--points-at", "HEAD"]).decode().splitlines()
    if git_tag not in git_tags:
        sys.exit("Could not find correspoing git tag: " + git_tag)

    # TO DO Check that version does not exist in PyPI


def get_readme_description():
    with open(join(ROOT, "readme.md")) as file_handle:
        return file_handle.read()


def get_package_requirements():
    requirements = []
    with open(join(ROOT, "requirements.txt")) as file_handle:
        # Requirements file contains one package name per line
        for line_data in file_handle.readlines():
            if line_data:
                requirements.append(line_data.strip())

    return requirements


def get_package_data():
    """
    Additional files that shall be included in the release, apart from the python packages
    """
    package_data = []
    package_data += list(find_git_files(directory=join(ROOT, "tsfpga", "tcl")))
    package_data += list(find_git_files(directory=join(ROOT, "modules")))

    # Specify path relative to the tsfpga python package folder
    package_data = [relpath(file_name, join(ROOT, "tsfpga")) for file_name in package_data]
    return package_data


if __name__ == "__main__":
    main()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from collections import OrderedDict
from os.path import dirname, join
import pytest
import unittest

from tsfpga.build_step_tcl_hook import BuildStepTclHook
from tsfpga.module import get_modules
from tsfpga.system_utils import create_file, delete
from tsfpga.vivado_tcl import VivadoTcl
from tsfpga.vivado_utils import to_tcl_path


THIS_DIR = dirname(__file__)


class TestVivadoTcl(unittest.TestCase):  # pylint: disable=too-many-instance-attributes

    modules_folder = join(THIS_DIR, "modules")

    def setUp(self):
        delete(self.modules_folder)

        # A library with some synth files and some test files
        self.a_vhd = to_tcl_path(create_file(join(self.modules_folder, "apa", "a.vhd")))
        self.tb_a_vhd = to_tcl_path(create_file(join(self.modules_folder, "apa", "test", "tb_a.vhd")))
        self.a_xdc = to_tcl_path(create_file(join(self.modules_folder, "apa", "scoped_constraints", "a.xdc")))

        self.b_v = to_tcl_path(create_file(join(self.modules_folder, "apa", "b.v")))
        self.b_tcl = to_tcl_path(create_file(join(self.modules_folder, "apa", "scoped_constraints", "b.tcl")))

        self.c_tcl = to_tcl_path(create_file(join(self.modules_folder, "apa", "ip_cores", "c.tcl")))

        # A library with only test files
        self.c_vhd = to_tcl_path(create_file(join(self.modules_folder, "zebra", "test", "c.vhd")))

        self.modules = get_modules([self.modules_folder])

        self.tcl = VivadoTcl(name="name")

    def test_only_synthesis_files_added_to_create_project_tcl(self):
        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="",
            top=""
        )
        assert self.a_vhd in tcl and self.b_v in tcl
        assert self.tb_a_vhd not in tcl and "tb_a.vhd" not in tcl

    def test_different_hdl_file_types(self):
        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="",
            top=""
        )
        assert f"read_vhdl -library apa -vhdl2008 {{{self.a_vhd}}}" in tcl
        assert f"read_verilog {{{self.b_v}}}" in tcl

    def test_empty_library_not_in_create_project_tcl(self):
        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="",
            top=""
        )
        assert "zebra" not in tcl

    def test_static_generics(self):
        # Use OrderedDict here in test so that order will be preserved and we can test for equality.
        # In real world case a normal dict can be used.
        generics = OrderedDict(enable=True, disable=False, integer=123, slv="4'b0101")

        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="part",
            top="",
            generics=generics
        )
        expected = "\nset_property generic {enable=1'b1 disable=1'b0 integer=123 slv=4'b0101} [current_fileset]\n"
        assert expected in tcl

    def test_constraints(self):
        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="part",
            top=""
        )

        expected = "\nread_xdc -ref a %s\n" % self.a_xdc
        assert expected in tcl
        expected = "\nread_xdc -ref b -unmanaged %s\n" % self.b_tcl
        assert expected in tcl

    def test_multiple_tcl_sources(self):
        extra_tcl_sources = ["dummy.tcl", "files.tcl"]
        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="part",
            top="",
            tcl_sources=extra_tcl_sources
        )

        for filename in extra_tcl_sources:
            assert f"\nsource -notrace {to_tcl_path(filename)}\n" in tcl

    def test_build_step_hooks(self):
        dummy = BuildStepTclHook("dummy.tcl", "STEPS.SYNTH_DESIGN.TCL.PRE")
        files = BuildStepTclHook("files.tcl", "STEPS.ROUTE_DESIGN.TCL.PRE")
        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="part",
            top="",
            build_step_hooks=[dummy, files]
        )

        assert f"\nset_property STEPS.SYNTH_DESIGN.TCL.PRE {to_tcl_path(dummy.tcl_file)} [get_runs synth_1]\n" in tcl
        assert f"\nset_property STEPS.ROUTE_DESIGN.TCL.PRE {to_tcl_path(files.tcl_file)} [get_runs impl_1]\n" in tcl

    def test_build_step_hooks_with_same_hook_step_should_raise_exception(self):
        dummy = BuildStepTclHook("dummy.tcl", "STEPS.SYNTH_DESIGN.TCL.PRE")
        files = BuildStepTclHook("files.tcl", "STEPS.SYNTH_DESIGN.TCL.PRE")
        with pytest.raises(ValueError) as exception_info:
            self.tcl.create(
                project_folder="",
                modules=self.modules,
                part="part",
                top="",
                build_step_hooks=[dummy, files]
            )

        assert str(exception_info.value).startswith("Multiple TCL sources for hook step STEPS.SYNTH_DESIGN.TCL.PRE")

    def test_ip_core_files(self):
        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="part",
            top=""
        )
        assert "\nsource -notrace %s\n" % self.c_tcl in tcl

    def test_ip_cache_location(self):
        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="part",
            top="",
        )
        assert "config_ip_cache" not in tcl

        ip_cache_location = to_tcl_path(THIS_DIR)
        tcl = self.tcl.create(
            project_folder="",
            modules=self.modules,
            part="part",
            top="",
            ip_cache_path=ip_cache_location
        )
        assert f"\nconfig_ip_cache -use_cache_location {ip_cache_location}\n" in tcl

    def test_set_multiple_threads(self):
        num_threads = 2
        tcl = self.tcl.build(
            project_file="",
            output_path="",
            num_threads=num_threads
        )
        assert "set_param general.maxThreads %d" % num_threads in tcl
        assert "launch_runs synth_1 -jobs %d" % num_threads in tcl
        assert "launch_runs impl_1 -jobs %d" % num_threads in tcl

    def test_runtime_generics(self):
        generics = dict(dummy=True)
        tcl = self.tcl.build(
            project_file="",
            output_path="",
            num_threads=0,
            generics=generics
        )
        expected = "\nset_property generic {dummy=1'b1} [current_fileset]\n"
        assert expected in tcl



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import join, dirname
import pytest
import subprocess
import sys
import unittest

from tsfpga.git_utils import find_git_files
from tsfpga.system_utils import create_file, delete


THIS_DIR = dirname(__file__)


def run_pylint(files):
    config = join(THIS_DIR, "pylintrc")
    command = [sys.executable, "-m", "pylint", "--rcfile=" + config] + files

    subprocess.check_call(command)


def test_pylint():
    files = list(find_git_files(file_ending="py"))
    run_pylint(files)


def run_pycodestyle(files):
    config = join(THIS_DIR, "pycodestylerc")
    command = [sys.executable, "-m", "pycodestyle", "--config=" + config] + files

    subprocess.check_call(command)


def test_pycodestyle():
    files = list(find_git_files(file_ending="py"))
    run_pycodestyle(files)


class TestPythonLintFunctions(unittest.TestCase):

    file = join(THIS_DIR, "dummy_python_file.py")
    ugly_code = """
aa  =
def bb:
cc  = 3
"""

    def setUp(self):
        delete(self.file)
        create_file(self.file, self.ugly_code)

    def test_pylint_should_raise_exception_if_there_are_ugly_files(self):
        with pytest.raises(subprocess.CalledProcessError):
            run_pylint([self.file])

    def test_pycodestyle_should_raise_exception_if_there_are_ugly_files(self):
        with pytest.raises(subprocess.CalledProcessError):
            run_pycodestyle([self.file])



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os import makedirs
from os.path import join, exists

from tsfpga import TSFPGA_TCL
from tsfpga.vivado_tcl import VivadoTcl
from tsfpga.vivado_utils import run_vivado_tcl


class VivadoProject:
    """
    Used for handling a Xilinx Vivado HDL project
    """

    # pylint: disable=too-many-arguments,too-many-instance-attributes
    def __init__(
            self,
            name,
            modules,
            part,
            top=None,
            generics=None,
            constraints=None,
            tcl_sources=None,
            build_step_hooks=None,
            vivado_path=None,
            defined_at=None,
    ):
        """
        :param name: Project name string.
        :param modules: A list of Module objects.
        :param part: Part identification string.
        :param top: Name of top level entity.
        :param generics: A dict with generics values. Use for static generics that
                         do not change between multiple builds of this project.
        :param constraints: A list of Constraint objects.
        :param tcl_sources: A list of paths to TCL files. Use for e.g. block design, settings, etc.
        :param build_step_hooks: A list of BuildStepTclHook objects.
        :param vivado_path: Default: Whatever version/location is in PATH will be used.
        :param defined_at: Optional string. To get a useful --list message.
        """
        self.name = name
        self.modules = modules
        self.part = part
        self.static_generics = generics
        self.defined_at = defined_at
        self.build_step_hooks = build_step_hooks

        self.top = name + "_top" if top is None else top
        self.vivado_path = "vivado" if vivado_path is None else vivado_path

        self._setup_constraints_list(constraints)
        self._setup_tcl_sources_list(tcl_sources)

        self.tcl = VivadoTcl(name=self.name)

    def _setup_constraints_list(self, constraints_from_user):
        # Lists are imutable. Since we assign and modify this one we have to copy it.
        self.constraints = None if constraints_from_user is None else constraints_from_user.copy()

    def _setup_tcl_sources_list(self, tcl_sources_from_user):
        # Lists are imutable. Since we assign and modify this one we have to copy it.
        self.tcl_sources = [] if tcl_sources_from_user is None else tcl_sources_from_user.copy()
        self.tcl_sources += [
            join(TSFPGA_TCL, "vivado_directive_settings.tcl"),
            join(TSFPGA_TCL, "vivado_messages.tcl"),
        ]

    def project_file(self, project_path):
        """
        Return the project file path of this project, in the given folder
        """
        return join(project_path, self.name + ".xpr")

    def _create_tcl(self, project_path, ip_cache_path):
        """
        Make a TCL file that creates a Vivado project
        """
        if exists(project_path):
            raise ValueError("Folder already exists: " + project_path)
        makedirs(project_path)

        create_vivado_project_tcl = join(project_path, "create_vivado_project.tcl")
        with open(create_vivado_project_tcl, "w") as file_handle:
            file_handle.write(self.tcl.create(
                project_folder=project_path,
                modules=self.modules,
                part=self.part,
                top=self.top,
                generics=self.static_generics,
                constraints=self.constraints,
                tcl_sources=self.tcl_sources,
                build_step_hooks=self.build_step_hooks,
                ip_cache_path=ip_cache_path,
            ))

        return create_vivado_project_tcl

    def create(self, project_path, ip_cache_path=None):
        """
        Create a Vivado project
        """
        print("Creating Vivado project in " + project_path)
        create_vivado_project_tcl = self._create_tcl(project_path, ip_cache_path)
        run_vivado_tcl(self.vivado_path, create_vivado_project_tcl)

    def _build_tcl(self, project_path, output_path, generics, synth_only, num_threads):
        """
        Make a TCL file that builds a Vivado project
        """
        project_file = self.project_file(project_path)
        if not exists(project_file):
            raise ValueError("Project file does not exist in the specified location: " + project_file)

        if self.static_generics is None:
            all_generics = generics
        else:
            # Add the two dictionaries
            all_generics = self.static_generics.copy()
            all_generics.update(generics)

        build_vivado_project_tcl = join(project_path, "build_vivado_project.tcl")
        with open(build_vivado_project_tcl, "w") as file_handle:
            file_handle.write(self.tcl.build(
                project_file=project_file,
                output_path=output_path,
                num_threads=num_threads,
                generics=all_generics,
                synth_only=synth_only))

        return build_vivado_project_tcl

    def pre_build(self, **kwargs):
        """
        Override this function in a child class if you wish to do something useful with it.
        """

    def post_build(self, **kwargs):
        """
        Override this function in a child class if you wish to do something useful with it.
        """

    def build(self,
              project_path,
              output_path=None,
              generics=None,
              synth_only=False,
              num_threads=12,
              **pre_and_post_build_parameters):
        """
        Build a Vivado project

        :param project_path: A path containing a Vivado project.
        :param output_path: Results (bit file, ...) will be placed here.
        :param generics: A dict with generics values. Use for run-time generics.
                         Values that can change between each build of this project.
        :param synth_only: Run synthesis and then stop.
        :param num_threads: Number of parallell threads to use during run.
        :param pre_and_post_build_parameters: Additional parameters that will be
                                              sent to pre- and post build (kwargs).
        """
        if output_path is None and not synth_only:
            raise ValueError("Must specify output_path when doing an implementation run")

        if synth_only:
            print("Synthesizing Vivado project in " + project_path)
        else:
            print(f"Building Vivado project in {project_path}, placing artifacts in {output_path}")

        # Send all available information to pre- and post build
        pre_and_post_build_parameters.update(
            project_path=project_path,
            output_path=output_path,
            generics=generics,
            synth_only=synth_only,
            num_threads=num_threads
        )

        self.pre_build(**pre_and_post_build_parameters)

        build_vivado_project_tcl = self._build_tcl(project_path=project_path,
                                                   output_path=output_path,
                                                   synth_only=synth_only,
                                                   generics=generics,
                                                   num_threads=num_threads)
        run_vivado_tcl(self.vivado_path, build_vivado_project_tcl)

        self.post_build(**pre_and_post_build_parameters)

    def __str__(self):
        result = str(self.__class__.__name__)
        if self.defined_at is not None:
            result += " defined at: " + self.defined_at
        result += "\nName:      " + self.name
        result += "\nTop level: " + self.top
        if self.static_generics is None:
            generics = "-"
        else:
            generics = ", ".join([f"{name}={value}" for name, value in self.static_generics.items()])
        result += "\nGenerics:  " + generics

        return result



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from collections import OrderedDict
import json
from os.path import join

from tsfpga.register_c_generator import RegisterCGenerator
from tsfpga.register_cpp_generator import RegisterCppGenerator
from tsfpga.register_html_generator import RegisterHtmlGenerator
from tsfpga.register_list import RegisterList, Register
from tsfpga.register_vhdl_generator import RegisterVhdlGenerator
from tsfpga.system_utils import create_file


def get_default_registers():
    registers = OrderedDict(
        config=Register("config", 0, "r_w", "Configuration register."),
        command=Register(
            "command", 1, "wpulse",
            "When this register is written, all '1's in the written word will be asserted for one "
            "clock cycle in the FPGA logic."),
        status=Register("status", 2, "r", "Status register."),
        irq_status=Register(
            "irq_status", 3, "r_wpulse",
            "Reading a '1' in this register means the corresponding interrupt has triggered. Writing "
            "to this register will clear the interrupts where there is a '1' in the written word."),
        irq_mask=Register(
            "irq_mask", 4, "r_w",
            "A '1' in this register means that the corresponding interrupt is enabled. ")
    )
    return registers


class Registers:

    def __init__(self, register_list):
        self.register_list = register_list

    def create_vhdl_package(self, output_path):
        """
        Assumes that the output_path folder already exists.
        This assumption makes it slightly faster than the other functions that use create_file().
        Necessary since this one is often used in real time (before simulations, etc..) and not in
        one-off scenarios like the others (when making a release).
        """
        with open(join(output_path, self.register_list.name + "_regs_pkg.vhd"), "w") as file_handle:
            file_handle.write(RegisterVhdlGenerator(self.register_list).get_package())

    def create_c_header(self, output_path):
        output_file = join(output_path, self.register_list.name + "_regs.h")
        create_file(output_file, RegisterCGenerator(self.register_list).get_header())

    def create_cpp_interface(self, output_path):
        output_file = join(output_path, "i_" + self.register_list.name + ".h")
        create_file(output_file, RegisterCppGenerator(self.register_list).get_interface())

    def create_cpp_header(self, output_path):
        output_file = join(output_path, self.register_list.name + ".h")
        create_file(output_file, RegisterCppGenerator(self.register_list).get_header())

    def create_cpp_implementation(self, output_path):
        output_file = join(output_path, self.register_list.name + ".cpp")
        create_file(output_file, RegisterCppGenerator(self.register_list).get_implementation())

    def create_html_page(self, output_path):
        output_file = join(output_path, self.register_list.name + "_regs.html")
        create_file(output_file, RegisterHtmlGenerator(self.register_list).get_page())

    def create_html_table(self, output_path):
        output_file = join(output_path, self.register_list.name + "_regs_table.html")
        create_file(output_file, RegisterHtmlGenerator(self.register_list).get_table())


def load_json_file(file_name):
    def check_for_duplicate_keys(ordered_pairs):
        """
        Raise ValueError if a duplicate key exists
        Note that built in dictionaries are not ordered, hence OrderedDict
        """
        result = OrderedDict()
        for key, value in ordered_pairs:
            if key in result:
                raise ValueError(f"Duplicate key {key}")
            result[key] = value
        return result

    try:
        with open(file_name) as file_handle:
            return json.load(file_handle, object_pairs_hook=check_for_duplicate_keys)
    except ValueError as exception_info:
        message = f"Error while parsing JSON file {file_name}:\n{exception_info}"
        raise ValueError(message)
    except FileNotFoundError:
        raise FileNotFoundError(f"Requested json file does not exist: {file_name}")


def from_json(module_name, json_file, default_registers=None):
    json_data = load_json_file(json_file)
    register_list = RegisterList(module_name)
    if default_registers is not None:
        register_list.registers = default_registers

    for register_name, register_fields in json_data.items():
        if default_registers is not None and register_name in default_registers:
            # Default registers can be "updated" in the sense that the user can use a custom
            # description and add whatever bits they use in the current module. They can not however
            # change the mode.
            register = register_list.registers[register_name]
            if "mode" in register_fields:
                mesg = f"Overloading register {register_name} in {json_file}, one can not change mode from default"
                raise ValueError(mesg)
        else:
            # If it is a new register however the mode has to be specified.
            if "mode" not in register_fields:
                raise ValueError(f"Register {register_name} in {json_file} does not have mode field")
            register = register_list.append(register_name, register_fields["mode"])

        if "description" in register_fields:
            register.description = register_fields["description"]

        if "bits" in register_fields:
            for bit_name, bit_description in register_fields["bits"].items():
                register.append_bit(bit_name, bit_description)

    return Registers(register_list)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.register_list import REGISTER_MODES
from tsfpga.markdown_to_html_translator import MarkdownToHtmlTranslator


class RegisterHtmlGenerator:

    def __init__(self, register_list):
        self.register_list = register_list
        self._markdown_to_html = MarkdownToHtmlTranslator()

    @staticmethod
    def _comment(comment):
        return f"<!-- {comment} -->\n"

    def _header(self):
        html = self._comment(self.register_list.generated_info())
        html += self._comment(self.register_list.generated_source_info())
        return html

    def _annotate_register(self, register):
        description = self._markdown_to_html.translate(register.description)
        html = f"""
  <tr>
    <td><strong>{register.name}</strong></td>
    <td>{register.address}</td>
    <td>{register.mode_readable}</td>
    <td>{description}</td>
  </tr>"""

        return html

    def _annotate_bit(self, bit):
        description = self._markdown_to_html.translate(bit.description)
        html = f"""
  <tr>
    <td>&nbsp;&nbsp;<em>{bit.name}</em></td>
    <td>{bit.idx}</td>
    <td></td>
    <td>{description}</td>
  </tr>"""

        return html

    def _get_table(self):
        html = """
<table>
<thead>
  <tr>
    <th>Name</th>
    <th>Address</th>
    <th>Mode</th>
    <th>Description</th>
  </tr>
</thead>
<tbody>"""

        for register in self.register_list.iterate_registers():
            html += self._annotate_register(register)
            for bit in register.bits:
                html += self._annotate_bit(bit)
        html += """
</tbody>
</table>"""

        return html

    def get_table(self):
        html = self._header()
        html += self._get_table()
        return html

    @staticmethod
    def _get_mode_descriptions():
        html = """
<table>
<thead>
  <tr>
    <th>Mode</th>
    <th>Description</th>
  </tr>
</thead>
<tbody>"""

        for (mode_readable, description) in REGISTER_MODES.values():
            html += f"""
<tr>
  <td>{mode_readable}</td>
  <td>{description}</td>
</tr>
"""
        html += """
</tbody>
</table>"""
        return html

    def get_page(self, table_style=None, font_style=None, extra_style=""):
        module_name = self.register_list.name
        title = f"Documentation of {module_name} registers"

        if font_style is None:
            font_style = """
html * {
  font-family: "Trebuchet MS", Arial, Helvetica, sans-serif;
}"""

        if table_style is None:
            table_style = """
table {
  border-collapse: collapse;
}
td, th {
  border: 1px solid #ddd;
  padding: 8px;
}
tr:nth-child(even) {
  background-color: #f2f2f2;
}
tr:hover {
  background-color: #ddd;
}
th {
  padding-top: 12px;
  padding-bottom: 12px;
  text-align: left;
  background-color: #4CAF50;
  color: white;
}"""

        footer = self.register_list.generated_source_info()
        html = f"""
{self._header()}

<!DOCTYPE html>
<html>
<head>
  <title>{title}</title>
  <style>
{font_style}
{table_style}
{extra_style}
  </style>
</head>
<body>
  <h1>{title}</h1>
  <p>This document is a specification of the PS interface of the {module_name} module.</p>
  <h2>Register modes</h2>
  <p>The following register modes are available.</p>
{self._get_mode_descriptions()}
  <h2>Register map</h2>
  <p>The following registers make up the register map for the {module_name} module.</p>
{self._get_table()}
<p>{footer}</p>
</body>
</html>"""

        return html



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


class HdlFile:

    vhdl_file_ending = ".vhd"
    verilog_source_file_ending = ".v"
    verilog_header_file_ending = ".vh"
    file_endings = (vhdl_file_ending, verilog_source_file_ending, verilog_header_file_ending)

    def __init__(self, filename):
        self.filename = filename

    @property
    def is_vhdl(self):
        return self.filename.endswith(self.vhdl_file_ending)

    @property
    def is_verilog_source(self):
        return self.filename.endswith(self.verilog_source_file_ending)

    @property
    def is_verilog_header(self):
        return self.filename.endswith(self.verilog_header_file_ending)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


class RegisterCppGenerator:

    def __init__(self, register_list):
        self.register_list = register_list
        self._class_name = self._to_pascal_case(register_list.name)

    def get_interface(self):
        cpp_code = f"class I{self._class_name}\n"
        cpp_code += "{\n"
        cpp_code += "public:\n"

        for register in self.register_list.registers.values():
            bits = register.bits
            if bits:
                for bit in bits:
                    bit_constant_name = register.name + "_" + bit.name
                    bit_constant_value = 2 ** bit.idx
                    cpp_code += f"  static const uint32_t {bit_constant_name} = {bit_constant_value}uL;\n"
                cpp_code += "\n"

        cpp_code += f"  virtual ~I{self._class_name}() {{ }}\n\n"
        for register in self.register_list.registers.values():
            if register.is_ps_readable:
                cpp_code += f"  virtual {self._getter_function_signature(register)} const = 0;\n"
            if register.is_ps_writeable:
                cpp_code += f"  virtual {self._setter_function_signature(register)} const = 0;\n"
            cpp_code += "\n"
        cpp_code += "};\n"

        cpp_code_top = f"""
{self._file_header()}
#pragma once

#include <cstdint>

"""
        return cpp_code_top + self._with_namespace(cpp_code)

    def get_header(self):
        cpp_code = f"class {self._class_name} : public I{self._class_name}\n"
        cpp_code += "{\n"

        cpp_code += "private:\n"
        cpp_code += "  volatile uint32_t *m_registers;\n\n"

        cpp_code += "public:\n"
        cpp_code += f"  {self._constructor_signature()};\n\n"
        cpp_code += f"  virtual ~{self._class_name}() {{ }}\n\n"
        for register in self.register_list.registers.values():
            if register.is_ps_readable:
                cpp_code += f"  virtual {self._getter_function_signature(register)} const override;\n"
            if register.is_ps_writeable:
                cpp_code += f"  virtual {self._setter_function_signature(register)} const override;\n"
            cpp_code += "\n"
        cpp_code += "};\n"

        cpp_code_top = f"""
{self._file_header()}
#pragma once

#include "i_{self.register_list.name}.h"

"""
        return cpp_code_top + self._with_namespace(cpp_code)

    def get_implementation(self):
        cpp_code = f"{self._class_name}::{self._constructor_signature()}\n"
        cpp_code += "    : m_registers(reinterpret_cast<volatile uint32_t *>(base_address))\n"
        cpp_code += "{\n"
        cpp_code += "  // Empty\n"
        cpp_code += "}\n\n"

        for register in self.register_list.registers.values():
            if register.is_ps_readable:
                cpp_code += self._getter_function(register)
            if register.is_ps_writeable:
                cpp_code += self._setter_function(register)

        cpp_code_top = f"{self._file_header()}\n"
        cpp_code_top += f"#include \"include/{self.register_list.name}.h\"\n\n"

        return cpp_code_top + self._with_namespace(cpp_code)

    @staticmethod
    def _with_namespace(cpp_code_body):
        cpp_code = "namespace fpga_regs\n"
        cpp_code += "{\n\n"
        cpp_code += f"{cpp_code_body}"
        cpp_code += "\n} /* namespace fpga_regs */\n"
        return cpp_code

    @staticmethod
    def _comment(comment):
        return f"/* {comment} */\n"

    def _file_header(self):
        cpp_code = self._comment(self.register_list.generated_info())
        cpp_code += self._comment(self.register_list.generated_source_info())
        return cpp_code

    @staticmethod
    def _to_pascal_case(snake_string):
        """
        Returns e.g., my_funny_string -> MyFunnyString
        """
        return snake_string.title().replace("_", "")

    def _constructor_signature(self):
        return f"{self._class_name}(volatile uint8_t *base_address)"

    def _setter_function(self, register):
        cpp_code = f"void {self._class_name}::set_{register.name}(uint32_t value) const\n"
        cpp_code += "{\n"
        cpp_code += f"  m_registers[{register.idx}] = value;\n"
        cpp_code += "}\n\n"
        return cpp_code

    @staticmethod
    def _setter_function_signature(register):
        return f"void set_{register.name}(uint32_t value)"

    def _getter_function(self, register):
        cpp_code = f"uint32_t {self._class_name}::get_{register.name}() const\n"
        cpp_code += "{\n"
        cpp_code += f"  return m_registers[{register.idx}];\n"
        cpp_code += "}\n\n"
        return cpp_code

    @staticmethod
    def _getter_function_signature(register):
        return f"uint32_t get_{register.name}()"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


class RegisterCGenerator:

    def __init__(self, register_list):
        self.register_list = register_list

    @staticmethod
    def _comment(comment):
        return f"/* {comment} */\n"

    def _header(self):
        c_code = self._comment(self.register_list.generated_info())
        c_code += self._comment(self.register_list.generated_source_info())
        return c_code

    def _register_struct(self):
        c_code = f"struct {self.register_list.name}_regs_t {{\n"
        for register in self.register_list.iterate_registers():
            c_code += f"  uint32_t {register.name};\n"
        c_code += "};\n"
        return c_code

    def _register_bits(self):
        c_code = ""
        for register in self.register_list.iterate_registers():
            for bit in register.bits:
                bit_name = (self.register_list.name + "_" + register.name + "_" + bit.name).upper()
                c_code += f"#define {bit_name}_BIT ({bit.idx}uL)\n"
                c_code += f"#define {bit_name}     ({2**bit.idx}uL)\n"
                c_code += "\n"

        return c_code

    def get_header(self):
        define_name = self.register_list.name.upper() + "_REGS_H"

        c_code = f"""
{self._header()}

#ifndef {define_name}
#define {define_name}
#pragma pack(push, 1)

#include <stdint.h>

{self._register_struct()}
{self._register_bits()}

#pragma pack(pop)
#endif {self._comment(define_name)}"""

        return c_code
